# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, List, Tuple, Union

import torch.utils.data
from lhotse.cut import Cut, CutSet
from lhotse.dataset import AudioSamples
from lhotse.dataset.collation import collate_vectors as collate_vectors_lhotse

from nemo.collections.common.data.lhotse.text_adapters import AudioTurn, NeMoMultimodalConversation, TextTurn
from nemo.collections.speechlm.data.dataset.data_utils import build_loss_mask
from nemo.collections.speechlm.data.text_processing import MultimodalConversationTextProcessor, TextProcessorOutput


def collate_vectors(items, max_length: int, padding_value):
    vectors = collate_vectors_lhotse(items, padding_value=padding_value)
    if max_length > vectors.size(1):
        vectors = torch.cat(
            [vectors, padding_value * torch.ones(vectors.size(0), max_length - vectors.size(1), dtype=vectors.dtype)],
            dim=1,
        )
    if items[0].shape[0] < 1:
        vectors = vectors.long()
    return vectors


class MultimodalConversationDataset(torch.utils.data.Dataset):
    """
    This dataset is based on Lhotse ASR dataset from ``audio_to_text_lhotse.py``
    and ``TarredAudioQuestionAnswerDataset`` from ``audio_text_qa_dataset.py``.

    Unlike native NeMo datasets, Lhotse dataset defines only the mapping from
    a CutSet (meta-data) to a mini-batch with PyTorch tensors.
    Specifically, it performs tokenization, I/O, augmentation, and feature extraction (if any).
    Managing data, sampling, de-duplication across workers/nodes etc. is all handled
    by Lhotse samplers instead.

    Args:
        text_processor: TextProcessing object
        default_context: Default question to use if no question is provided
        tokens_to_generate: Number of tokens to generate during inference
        pad_to_max_length: Whether to pad the input to the max sequence length. If False, will pad to the max length of the current batch.
        max_seq_length: Maximum sequence length for each dataset examples. Examples will either be truncated to fit this length or dropped if they cannot be truncated.
        context_key: Key to use for the context in your JSONL file
        default_context_key: Key to use for the default context in lhotse yaml
    """

    def __init__(
        self,
        text_processor: MultimodalConversationTextProcessor,
        tokens_to_generate: int,
        pad_to_max_length: bool,
        max_seq_length: int,
        default_context: str = "listen to the audio",
        context_key: str = "context",
        default_context_key: str = "default_context",
        answer_key: str = "answer",
        answer_only_loss: bool = True,
        is_train: bool = False,
    ):
        super().__init__()
        self.text_processor = text_processor
        self.load_audio = AudioSamples(fault_tolerant=True)
        self.tokens_to_generate = tokens_to_generate
        self.pad_to_max_length = pad_to_max_length
        self.max_seq_length = max_seq_length

        self.default_context = default_context
        self.context_key = context_key
        self.default_context_key = default_context_key
        self.answer_key = answer_key
        self.answer_only_loss = answer_only_loss
        self.is_train = is_train

    def __getitem__(self, all_cuts: CutSet) -> dict[str, Union[torch.Tensor, list[str], dict]]:
        audio_samples = []
        text_samples = []
        for sample in all_cuts:
            audio_data, text_data = self._process_sample(sample)
            audio_samples.append(audio_data)
            text_samples.append(text_data)

        audio_batch = collate_audio_data(audio_samples, pad_val=0.0)

        text_batch = collate_text_data(
            text_samples,
            self.tokens_to_generate,
            self.pad_to_max_length,
            self.max_seq_length,
            pad_id=self.text_processor.pad_id,
            answer_only_loss=self.answer_only_loss,
        )

        sample_ids = list(all_cuts.ids)
        metadata = self._get_metadata(all_cuts)
        batch = {
            "sample_ids": sample_ids,
            "metadata": metadata,
            "audio_signal": audio_batch["audio_signal"],
            "audio_signal_length": audio_batch["audio_length"],
            "tokens": text_batch["tokens"],
            "tokens_length": text_batch["tokens_length"],
            "labels": text_batch["labels"],
            "loss_mask": text_batch["loss_mask"],
            "position_ids": text_batch["position_ids"],
            "contexts": text_batch["contexts"],
            "context_lengths": text_batch["context_lengths"],
            "answers": text_batch["answers"],
            "max_length": text_batch["max_length"],
            "context_start_idx": text_batch["context_start_idx"],  # used for multi-audio per sample
            "num_audios": text_batch["num_audios"],  # used for multi-audio per sample
        }

        if self.is_train:
            # drop the context and answer that are not used in training
            batch.pop("contexts")
            batch.pop("context_lengths")
            batch.pop("answers")

        return batch

    def _get_metadata(self, all_cuts: CutSet) -> List[dict]:
        metadata = []
        for cut in all_cuts:
            metadata.append({"type": type(cut).__name__, "id": getattr(cut, "id", "n/a"), "cut": str(cut)})
        return metadata

    def _process_sample(self, sample: Any) -> dict:
        if isinstance(sample, Cut):
            return self._process_cut(sample)
        elif isinstance(sample, NeMoMultimodalConversation):
            return self._process_multimodal_conversation(sample)
        else:
            raise ValueError(f"Unsupported input type: {type(sample)}")

    def _convert_cut_sample(self, cut: Cut) -> NeMoMultimodalConversation:
        """
        Transform Cut input to as single turn MultiModalConversation for backward compatability.
        """
        if hasattr(cut, self.context_key):
            context = getattr(cut, self.context_key)
        elif hasattr(cut, self.default_context_key):
            context = getattr(cut, self.default_context_key)
        else:
            context = self.default_context

        if hasattr(cut, self.answer_key):
            answer = getattr(cut, self.answer_key)
        else:
            answer = cut.supervisions[0].text

        if context is None:
            raise ValueError(
                f"Context is None for cut {cut}, please double check the `context_key` and the actual content of that field in your data"
            )
        if answer is None:
            raise ValueError(
                f"Answer is None for cut {cut}, please double check the `answer_key` and the actual content of that field in your data"
            )

        sample = NeMoMultimodalConversation(
            id=cut.id,
            turns=[
                TextTurn(
                    role="user",
                    value=context,
                ),
                AudioTurn(
                    role="user",
                    audio_locator_tag="[audio]",
                    cut=cut,
                ),
                TextTurn(
                    role="assistant",
                    value=answer,
                ),
            ],
        )
        return sample

    def _process_cut(self, cut: Cut) -> Tuple[dict, TextProcessorOutput]:
        audio_signal = cut.load_audio().reshape(-1)
        audio_data = {
            "audio_signal": [audio_signal],
            "audio_length": [len(audio_signal)],
        }

        # Process text
        sample = self._convert_cut_sample(cut)
        text_data = self.text_processor(sample)

        return audio_data, text_data

    def _process_multimodal_conversation(self, sample: NeMoMultimodalConversation) -> Tuple[dict, TextProcessorOutput]:
        # Load audio
        audio_turns = [turn for turn in sample.turns if isinstance(turn, AudioTurn)]
        audio_signal = [turn.cut.load_audio().reshape(-1) for turn in audio_turns]
        audio_length = [len(audio) for audio in audio_signal]
        audio_data = {
            "audio_signal": audio_signal,
            "audio_length": audio_length,
        }

        # Process text
        text_data = self.text_processor(sample)

        return audio_data, text_data


def collate_audio_data(samples: List[dict], pad_val: int = 0) -> dict:
    """
    Collate audio data for all samples in the batch
    """
    audio_signal = []
    audio_length = []
    for sample in samples:
        audio_signal.extend(sample["audio_signal"])
        audio_length.extend(sample["audio_length"])

    if len(audio_signal) == 0:
        audio_signal = torch.tensor([0.0])
        audio_length = [0]

    max_audio_length = max(audio_length)
    audio_signal_tensor = collate_vectors(audio_signal, max_audio_length, pad_val)
    audio_length_tensor = torch.tensor(audio_length).long()
    return {
        "audio_signal": audio_signal_tensor,  # [batch_size, max_audio_length]
        "audio_length": audio_length_tensor,  # [batch_size]
    }


def collate_text_data(
    samples: List[TextProcessorOutput],
    tokens_to_generate: int,
    pad_to_max_length: bool,
    max_seq_length: int,
    pad_id: int,
    answer_only_loss: bool = True,
) -> dict:
    """
    Perform text collation on results from text processor
    """
    batch_size = len(samples)

    def get_max_len(input_list):
        return max([len(x) for x in input_list])

    input_ids = [sample.input_ids for sample in samples]
    context_ids = [sample.context_ids for sample in samples]
    context_lengths = [sample.context_length for sample in samples]
    answer_ids = [sample.answer_ids for sample in samples]
    context_start_idx = [sample.context_start_idx.numpy().tolist() for sample in samples]
    num_audios = [sample.num_audios for sample in samples]
    answer_start_idx = [sample.answer_start_idx for sample in samples]

    input_id_maxlen = get_max_len(input_ids)
    context_id_maxlen = tokens_to_generate + get_max_len(context_ids)
    answer_id_maxlen = get_max_len(answer_ids)
    if pad_to_max_length:
        input_id_maxlen = max_seq_length
        context_id_maxlen = max_seq_length
        answer_id_maxlen = max_seq_length

    all_tokens = collate_vectors(input_ids, max_length=input_id_maxlen, padding_value=pad_id)
    full_lengths = torch.LongTensor([len(item) for item in input_ids])

    assert input_id_maxlen <= max_seq_length, f"{input_id_maxlen=} <= {max_seq_length=}"

    loss_mask = [
        torch.as_tensor(build_loss_mask(inp_id, ans_idx, answer_only_loss))
        for inp_id, ans_idx in zip(input_ids, answer_start_idx)
    ]
    return {
        "tokens": all_tokens[:, :-1],  # [batch_size, input_id_maxlen - 1]
        "tokens_length": full_lengths - 1,  # [batch_size]
        "labels": all_tokens[:, 1:],  # [batch_size, input_id_maxlen - 1]
        "loss_mask": collate_vectors(loss_mask, max_length=input_id_maxlen, padding_value=0)[
            :, 1:
        ],  # [batch_size, input_id_maxlen - 1]
        "position_ids": torch.arange(input_id_maxlen, dtype=torch.long).repeat(
            batch_size, 1
        ),  # [batch_size, input_id_maxlen]
        "contexts": collate_vectors(
            context_ids, max_length=context_id_maxlen, padding_value=pad_id
        ),  # [batch_size, context_id_maxlen]
        "context_lengths": torch.LongTensor(context_lengths),  # [batch_size]
        "answers": collate_vectors(
            answer_ids, max_length=answer_id_maxlen, padding_value=pad_id
        ),  # [batch_size, answer_id_maxlen]
        "max_length": torch.LongTensor([input_id_maxlen] * batch_size),  # [batch_size]
        "context_start_idx": context_start_idx,  # List[List[int]],
        "num_audios": torch.stack(num_audios),  # torch.Tensor, shape (batch_size,)
    }



# Copyright (c) 2020, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nemo.collections.tts.models.aligner import AlignerModel
from nemo.collections.tts.models.audio_codec import AudioCodecModel
from nemo.collections.tts.models.fastpitch import FastPitchModel
from nemo.collections.tts.models.fastpitch_ssl import FastPitchModel_SSL
from nemo.collections.tts.models.hifigan import HifiGanModel
from nemo.collections.tts.models.magpietts import InferBatchOutput, MagpieTTSModel
from nemo.collections.tts.models.magpietts_preference_optimization import (
    MagpieTTSModelOfflinePO,
    MagpieTTSModelOfflinePODataGen,
    MagpieTTSModelOnlinePO,
)
from nemo.collections.tts.models.mixer_tts import MixerTTSModel
from nemo.collections.tts.models.radtts import RadTTSModel
from nemo.collections.tts.models.spectrogram_enhancer import SpectrogramEnhancerModel
from nemo.collections.tts.models.ssl_tts import SSLDisentangler
from nemo.collections.tts.models.tacotron2 import Tacotron2Model
from nemo.collections.tts.models.two_stages import GriffinLimModel, MelPsuedoInverseModel, TwoStagesModel
from nemo.collections.tts.models.univnet import UnivNetModel
from nemo.collections.tts.models.vits import VitsModel
from nemo.collections.tts.models.waveglow import WaveGlowModel

__all__ = [
    "AlignerModel",
    "AudioCodecModel",
    "FastPitchModel",
    "FastPitchModel_SSL",
    "SSLDisentangler",
    "GriffinLimModel",
    "HifiGanModel",
    "InferBatchOutput",
    "MelPsuedoInverseModel",
    "MixerTTSModel",
    "RadTTSModel",
    "MagpieTTSModel",
    "MagpieTTSModelOfflinePODataGen",
    "MagpieTTSModelOfflinePO",
    "MagpieTTSModelOnlinePO",
    "Tacotron2Model",
    "TwoStagesModel",
    "UnivNetModel",
    "VitsModel",
    "WaveGlowModel",
    "SpectrogramEnhancerModel",
]



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import contextlib
import copy
import os
import threading
from typing import Any, Dict, Iterable

import lightning.pytorch as pl
import torch
from lightning.pytorch import Callback
from lightning.pytorch.utilities.exceptions import MisconfigurationException
from lightning.pytorch.utilities.rank_zero import rank_zero_info


class EMA(Callback):
    """
    Implements Exponential Moving Averaging (EMA).

    When training a model, this callback will maintain moving averages of the trained parameters.
    When evaluating, we use the moving averages copy of the trained parameters.
    When saving, we save an additional set of parameters with the prefix `ema`.

    Args:
        decay: The exponential decay used when calculating the moving average. Has to be between 0-1.
        validate_original_weights: Validate the original weights, as apposed to the EMA weights.
        every_n_steps: Apply EMA every N steps.
        cpu_offload: Offload weights to CPU.
    """

    def __init__(
        self,
        decay: float,
        validate_original_weights: bool = False,
        every_n_steps: int = 1,
        cpu_offload: bool = False,
    ):
        if not (0 <= decay <= 1):
            raise MisconfigurationException("EMA decay value must be between 0 and 1")
        self.decay = decay
        self.validate_original_weights = validate_original_weights
        self.every_n_steps = every_n_steps
        self.cpu_offload = cpu_offload

    def on_fit_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        device = pl_module.device if not self.cpu_offload else torch.device('cpu')
        trainer.optimizers = [
            EMAOptimizer(
                optim,
                device=device,
                decay=self.decay,
                every_n_steps=self.every_n_steps,
                current_step=trainer.global_step,
            )
            for optim in trainer.optimizers
            if not isinstance(optim, EMAOptimizer)
        ]

    def on_validation_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        if self._should_validate_ema_weights(trainer):
            self.swap_model_weights(trainer)

    def on_validation_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        if self._should_validate_ema_weights(trainer):
            self.swap_model_weights(trainer)

    def on_test_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        if self._should_validate_ema_weights(trainer):
            self.swap_model_weights(trainer)

    def on_test_end(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule") -> None:
        if self._should_validate_ema_weights(trainer):
            self.swap_model_weights(trainer)

    def _should_validate_ema_weights(self, trainer: "pl.Trainer") -> bool:
        return not self.validate_original_weights and self._ema_initialized(trainer)

    def _ema_initialized(self, trainer: "pl.Trainer") -> bool:
        return any(isinstance(optimizer, EMAOptimizer) for optimizer in trainer.optimizers)

    def swap_model_weights(self, trainer: "pl.Trainer", saving_ema_model: bool = False):
        for optimizer in trainer.optimizers:
            assert isinstance(optimizer, EMAOptimizer)
            optimizer.switch_main_parameter_weights(saving_ema_model)

    @contextlib.contextmanager
    def save_ema_model(self, trainer: "pl.Trainer"):
        """
        Saves an EMA copy of the model + EMA optimizer states for resume.
        """
        self.swap_model_weights(trainer, saving_ema_model=True)
        try:
            yield
        finally:
            self.swap_model_weights(trainer, saving_ema_model=False)

    @contextlib.contextmanager
    def save_original_optimizer_state(self, trainer: "pl.Trainer"):
        for optimizer in trainer.optimizers:
            assert isinstance(optimizer, EMAOptimizer)
            optimizer.save_original_optimizer_state = True
        try:
            yield
        finally:
            for optimizer in trainer.optimizers:
                optimizer.save_original_optimizer_state = False

    def on_load_checkpoint(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule", checkpoint: Dict[str, Any]
    ) -> None:
        checkpoint_callback = trainer.checkpoint_callback

        # Replace connector._ckpt_path with below to avoid calling into lightning's protected API
        ckpt_path = trainer.ckpt_path

        if ckpt_path and checkpoint_callback is not None and 'NeMo' in type(checkpoint_callback).__name__:
            ext = checkpoint_callback.FILE_EXTENSION
            if ckpt_path.endswith(f'-EMA{ext}'):
                rank_zero_info(
                    "loading EMA based weights. "
                    "The callback will treat the loaded EMA weights as the main weights"
                    " and create a new EMA copy when training."
                )
                return
            ema_path = ckpt_path.replace(ext, f'-EMA{ext}')
            if os.path.exists(ema_path):
                ema_state_dict = torch.load(ema_path, map_location=torch.device('cpu'))

                checkpoint['optimizer_states'] = ema_state_dict['optimizer_states']
                del ema_state_dict
                rank_zero_info("EMA state has been restored.")
            else:
                raise MisconfigurationException(
                    "Unable to find the associated EMA weights when re-loading, "
                    f"training will start with new EMA weights. Expected them to be at: {ema_path}",
                )


@torch.no_grad()
def ema_update(ema_model_tuple, current_model_tuple, decay):
    torch._foreach_mul_(ema_model_tuple, decay)
    torch._foreach_add_(
        ema_model_tuple,
        current_model_tuple,
        alpha=(1.0 - decay),
    )


def run_ema_update_cpu(ema_model_tuple, current_model_tuple, decay, pre_sync_stream=None):
    if pre_sync_stream is not None:
        pre_sync_stream.synchronize()

    ema_update(ema_model_tuple, current_model_tuple, decay)


class EMAOptimizer(torch.optim.Optimizer):
    r"""
    EMAOptimizer is a wrapper for torch.optim.Optimizer that computes
    Exponential Moving Average of parameters registered in the optimizer.

    EMA parameters are automatically updated after every step of the optimizer
    with the following formula:

        ema_weight = decay * ema_weight + (1 - decay) * training_weight

    To access EMA parameters, use ``swap_ema_weights()`` context manager to
    perform a temporary in-place swap of regular parameters with EMA
    parameters.

    Notes:
        - EMAOptimizer is not compatible with APEX AMP O2.

    Args:
        optimizer (torch.optim.Optimizer): optimizer to wrap
        device (torch.device): device for EMA parameters
        decay (float): decay factor

    Returns:
        returns an instance of torch.optim.Optimizer that computes EMA of
        parameters

    Example:
        model = Model().to(device)
        opt = torch.optim.Adam(model.parameters())

        opt = EMAOptimizer(opt, device, 0.9999)

        for epoch in range(epochs):
            training_loop(model, opt)

            regular_eval_accuracy = evaluate(model)

            with opt.swap_ema_weights():
                ema_eval_accuracy = evaluate(model)
    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        device: torch.device,
        decay: float = 0.9999,
        every_n_steps: int = 1,
        current_step: int = 0,
    ):
        self.optimizer = optimizer
        self.decay = decay
        self.device = device
        self.current_step = current_step
        self.every_n_steps = every_n_steps
        self.save_original_optimizer_state = False

        self.first_iteration = True
        self.rebuild_ema_params = True
        self.stream = None
        self.thread = None

        self.ema_params = ()
        self.in_saving_ema_model_context = False

    def all_parameters(self) -> Iterable[torch.Tensor]:
        return (param for group in self.param_groups for param in group['params'])

    def step(self, closure=None, grad_scaler=None, **kwargs):
        self.join()

        if self.first_iteration:
            if any(p.is_cuda for p in self.all_parameters()):
                self.stream = torch.cuda.Stream()

            self.first_iteration = False

        if self.rebuild_ema_params:
            opt_params = list(self.all_parameters())

            self.ema_params += tuple(
                copy.deepcopy(param.data.detach()).to(self.device) for param in opt_params[len(self.ema_params) :]
            )
            self.rebuild_ema_params = False

        if getattr(self.optimizer, "_step_supports_amp_scaling", False) and grad_scaler is not None:
            loss = self.optimizer.step(closure=closure, grad_scaler=grad_scaler)
        else:
            loss = self.optimizer.step(closure)

        if self._should_update_at_step():
            self.update()
        self.current_step += 1
        return loss

    def _should_update_at_step(self) -> bool:
        return self.current_step % self.every_n_steps == 0

    @torch.no_grad()
    def update(self):
        if self.stream is not None:
            self.stream.wait_stream(torch.cuda.current_stream())

        with torch.cuda.stream(self.stream):
            current_model_state = tuple(
                param.data.to(self.device, non_blocking=True) for param in self.all_parameters()
            )

            if self.device.type == 'cuda':
                ema_update(self.ema_params, current_model_state, self.decay)

        if self.device.type == 'cpu':
            self.thread = threading.Thread(
                target=run_ema_update_cpu,
                args=(
                    self.ema_params,
                    current_model_state,
                    self.decay,
                    self.stream,
                ),
            )
            self.thread.start()

    def swap_tensors(self, tensor1, tensor2):
        tmp = torch.empty_like(tensor1)
        tmp.copy_(tensor1)
        tensor1.copy_(tensor2)
        tensor2.copy_(tmp)

    def switch_main_parameter_weights(self, saving_ema_model: bool = False):
        self.join()
        self.in_saving_ema_model_context = saving_ema_model
        for param, ema_param in zip(self.all_parameters(), self.ema_params):
            self.swap_tensors(param.data, ema_param)

    @contextlib.contextmanager
    def swap_ema_weights(self, enabled: bool = True):
        r"""
        A context manager to in-place swap regular parameters with EMA
        parameters.
        It swaps back to the original regular parameters on context manager
        exit.

        Args:
            enabled (bool): whether the swap should be performed
        """

        if enabled:
            self.switch_main_parameter_weights()
        try:
            yield
        finally:
            if enabled:
                self.switch_main_parameter_weights()

    def __getattr__(self, name):
        return getattr(self.optimizer, name)

    def join(self):
        if self.stream is not None:
            self.stream.synchronize()

        if self.thread is not None:
            self.thread.join()

    def state_dict(self):
        self.join()

        if self.save_original_optimizer_state:
            return self.optimizer.state_dict()

        # if we are in the context of saving an EMA model, the EMA weights are in the modules' actual weights
        ema_params = self.ema_params if not self.in_saving_ema_model_context else list(self.all_parameters())
        state_dict = {
            'opt': self.optimizer.state_dict(),
            'ema': ema_params,
            'current_step': self.current_step,
            'decay': self.decay,
            'every_n_steps': self.every_n_steps,
        }
        return state_dict

    def load_state_dict(self, state_dict):
        self.join()

        self.optimizer.load_state_dict(state_dict['opt'])
        self.ema_params = tuple(param.to(self.device) for param in copy.deepcopy(state_dict['ema']))
        self.current_step = state_dict['current_step']
        self.decay = state_dict['decay']
        self.every_n_steps = state_dict['every_n_steps']
        self.rebuild_ema_params = False

    def add_param_group(self, param_group):
        self.optimizer.add_param_group(param_group)
        self.rebuild_ema_params = True



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2018-2020 William Falcon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple

import torch

from .pl_utils import BATCH_SIZE, NUM_BATCHES, NUM_CLASSES

Input = namedtuple('Input', ["probs", "logits"])


ONLY_PROBS = Input(probs=torch.rand(NUM_BATCHES, BATCH_SIZE, NUM_CLASSES), logits=None)

ONLY_LOGITS1 = Input(probs=None, logits=torch.rand(NUM_BATCHES, BATCH_SIZE, NUM_CLASSES))

ONLY_LOGITS100 = Input(probs=None, logits=torch.rand(NUM_BATCHES, BATCH_SIZE, NUM_CLASSES) * 200 - 100)

PROBS_AND_LOGITS = Input(
    probs=torch.rand(NUM_BATCHES, BATCH_SIZE, NUM_CLASSES),
    logits=torch.rand(NUM_BATCHES, BATCH_SIZE, NUM_CLASSES) * 200 - 100,
)

NO_PROBS_NO_LOGITS = Input(probs=None, logits=None)



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import pytest
import torch

from nemo.collections.asr.modules.conformer_encoder import ConformerEncoder


class TestStochasticDepth:
    """Testing stochastic depth functionality."""

    def test_stochastic_depth_model_creation(self):
        """Testing basic model creation and the drop probs are correctly assigned."""
        n_layers = 4
        model = ConformerEncoder(feat_in=10, n_layers=n_layers, d_model=4, feat_out=8)

        # checking that by default SD is disabled
        assert model.layer_drop_probs == [0.0] * n_layers

        # linear mode
        for drop_prob in [0.3, 0.5, 0.9]:
            for start_layer in [1, 3]:
                model = ConformerEncoder(
                    feat_in=10,
                    n_layers=n_layers,
                    d_model=4,
                    feat_out=8,
                    stochastic_depth_drop_prob=drop_prob,
                    stochastic_depth_start_layer=start_layer,
                )
                L = n_layers - start_layer
                assert model.layer_drop_probs == [0.0] * start_layer + [drop_prob * l / L for l in range(1, L + 1)]

        # uniform mode
        for drop_prob in [0.3, 0.5, 0.9]:
            model = ConformerEncoder(
                feat_in=10,
                n_layers=n_layers,
                d_model=4,
                feat_out=8,
                stochastic_depth_drop_prob=drop_prob,
                stochastic_depth_mode="uniform",
                stochastic_depth_start_layer=start_layer,
            )
            L = n_layers - start_layer
            assert model.layer_drop_probs == [0.0] * start_layer + [drop_prob] * L

        # checking for errors
        for drop_prob in [-1.0, 1.0]:
            with pytest.raises(ValueError, match="stochastic_depth_drop_prob has to be in"):
                ConformerEncoder(
                    feat_in=10,
                    n_layers=n_layers,
                    d_model=4,
                    feat_out=8,
                    stochastic_depth_drop_prob=drop_prob,
                    stochastic_depth_mode="uniform",
                )

        with pytest.raises(ValueError, match="stochastic_depth_mode has to be one of"):
            ConformerEncoder(feat_in=10, n_layers=n_layers, d_model=4, feat_out=8, stochastic_depth_mode="weird")

        for start_layer in [-1, 0, 5]:
            with pytest.raises(ValueError, match="stochastic_depth_start_layer has to be in"):
                ConformerEncoder(
                    feat_in=10,
                    n_layers=n_layers,
                    d_model=4,
                    feat_out=8,
                    stochastic_depth_start_layer=start_layer,
                )

    @pytest.mark.pleasefixme
    def test_stochastic_depth_forward(self):
        """Testing that forward works and we get randomness during training, but not during eval."""
        random_input = torch.rand((1, 2, 2))
        random_length = torch.tensor([2], dtype=torch.int64)

        model = ConformerEncoder(
            feat_in=2,
            n_layers=3,
            d_model=4,
            feat_out=4,
            stochastic_depth_drop_prob=0.8,
            dropout=0.0,
            dropout_pre_encoder=0.0,
            dropout_emb=0.0,
            conv_norm_type="layer_norm",
            conv_kernel_size=3,
        )
        model.train()
        outputs = [None] * 5
        for i in range(5):
            outputs[i] = model(audio_signal=random_input, length=random_length)[0]
        # checking that not all outputs are the same
        num_diff = 0
        for i in range(1, 5):
            if not torch.allclose(outputs[i], outputs[0]):
                num_diff += 1
        assert num_diff > 0

        model.eval()
        outputs = [None] * 5
        for i in range(5):
            outputs[i] = model(audio_signal=random_input, length=random_length)[0]
        # checking that not all outputs are the same
        num_diff = 0
        for i in range(1, 5):
            if not torch.allclose(outputs[i], outputs[0]):
                num_diff += 1
        assert num_diff == 0


class TestBypassPreEncode:
    """Testing bypass pre-encode functionality."""

    def test_bypass_pre_encode_forward(self):
        """Testing that forward works with "bypass pre-encode" mode."""
        # For pre-encoded embeddings, the shape is (batch_size, n_frames, emb_dim)
        batch_size = 2
        n_frames, emb_dim, feat_out = 17, 16, 8
        random_input = torch.rand((batch_size, n_frames, emb_dim))
        random_length = torch.tensor([n_frames], dtype=torch.int64)

        model = ConformerEncoder(
            feat_in=10,
            n_layers=3,
            d_model=emb_dim,
            feat_out=feat_out,
            stochastic_depth_drop_prob=0.0,
            dropout=0.0,
            dropout_pre_encoder=0.0,
            dropout_emb=0.0,
            conv_norm_type="layer_norm",
            conv_kernel_size=3,
        )
        model.train()
        fwd_outputs = model(audio_signal=random_input, length=random_length, bypass_pre_encode=True)[0]
        assert fwd_outputs.shape == (batch_size, feat_out, n_frames)

        model.eval()
        fwd_outputs = model(audio_signal=random_input, length=random_length, bypass_pre_encode=True)[0]
        assert fwd_outputs.shape == (batch_size, feat_out, n_frames)

    def test_error_shape_invalid_bypass_pre_encode_forward(self):
        """
        Testing that error messages are correctly triggered regarding "bypass pre-encode" mode.
        Both correct samples and wrongs samples are tested.

        (1) bypass_pre_encode = False (default):
            `audio_signal` must be a tensor containing audio features.
            Shape: (batch, self._feat_in, n_frames)
        (2) bypass_pre_encode = True:
            `audio_signal` must be a tensor containing pre-encoded embeddings.
            Shape: (batch, n_frame, self.d_model)
        """
        batch_size = 2
        n_frames, emb_dim, feat_in, feat_out = 17, 16, 10, 8

        pre_encode_input = torch.rand((batch_size, n_frames, emb_dim))
        feat_input = torch.rand((batch_size, feat_in, n_frames))
        input_length = torch.tensor([n_frames], dtype=torch.int64)

        model = ConformerEncoder(
            feat_in=feat_in,
            n_layers=3,
            d_model=emb_dim,
            feat_out=feat_out,
            stochastic_depth_drop_prob=0.0,
            dropout=0.0,
            dropout_pre_encoder=0.0,
            dropout_emb=0.0,
            conv_norm_type="layer_norm",
            conv_kernel_size=3,
        )
        sub_sampled_n_frames = np.ceil(n_frames / model.subsampling_factor)

        # Test with bypass_pre_encode = True, should be pre_encode_input but given feat_input.
        model.train()
        with pytest.raises(ValueError):
            model(audio_signal=feat_input, length=input_length, bypass_pre_encode=True)

        model.eval()
        with pytest.raises(ValueError):
            model(audio_signal=feat_input, length=input_length, bypass_pre_encode=True)

        # Test with bypass_pre_encode = True, given the correct input pre_encode_input.
        model.train()
        fwd_outputs = model(audio_signal=pre_encode_input, length=input_length, bypass_pre_encode=True)[0]
        assert fwd_outputs.shape == (batch_size, feat_out, n_frames)

        model.eval()
        fwd_outputs = model(audio_signal=pre_encode_input, length=input_length, bypass_pre_encode=True)[0]
        assert fwd_outputs.shape == (batch_size, feat_out, n_frames)

        # Test with bypass_pre_encode = False, should be feat_input but given pre_encode_input.
        model.train()
        with pytest.raises(ValueError):
            model(audio_signal=pre_encode_input, length=input_length, bypass_pre_encode=False)

        model.eval()
        with pytest.raises(ValueError):
            model(audio_signal=pre_encode_input, length=input_length, bypass_pre_encode=False)

        # Test with bypass_pre_encode = False, given the correct input feat_input.
        model.train()
        fwd_outputs = model(audio_signal=feat_input, length=input_length, bypass_pre_encode=False)[0]
        assert fwd_outputs.shape == (batch_size, feat_out, sub_sampled_n_frames)

        model.eval()
        fwd_outputs = model(audio_signal=feat_input, length=input_length, bypass_pre_encode=False)[0]
        assert fwd_outputs.shape == (batch_size, feat_out, sub_sampled_n_frames)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional, Tuple

import torch.utils.data
from lhotse.dataset import AudioSamples
from lhotse.dataset.collation import collate_matrices

from nemo.collections.asr.parts.utils.asr_multispeaker_utils import (
    get_hidden_length_from_sample_length,
    speaker_to_target,
)
from nemo.core.neural_types import AudioSignal, LabelsType, LengthsType, NeuralType
from nemo.utils import logging


class LhotseAudioToSpeechE2ESpkDiarDataset(torch.utils.data.Dataset):
    """
    This dataset is a Lhotse version of diarization dataset in audio_to_diar_label.py.
    Unlike native NeMo datasets, Lhotse dataset defines only the mapping from
    a CutSet (meta-data) to a mini-batch with PyTorch tensors.
    Specifically, it performs tokenization, I/O, augmentation, and feature extraction (if any).
    Managing data, sampling, de-duplication across workers/nodes etc. is all handled
    by Lhotse samplers instead.
    """

    @property
    def output_types(self) -> Optional[Dict[str, NeuralType]]:
        """Define the output types of the dataset."""
        return {
            'audio_signal': NeuralType(('B', 'T'), AudioSignal()),
            'a_sig_length': NeuralType(tuple('B'), LengthsType()),
            'targets': NeuralType(('B', 'T', 'N'), LabelsType()),
            'target_length': NeuralType(tuple('B'), LengthsType()),
            'sample_id': NeuralType(tuple('B'), LengthsType(), optional=True),
        }

    def __init__(self, cfg):
        super().__init__()
        self.load_audio = AudioSamples(fault_tolerant=True)
        self.cfg = cfg
        self.num_speakers = self.cfg.get('num_speakers', 4)
        self.num_sample_per_mel_frame = int(
            self.cfg.get('window_stride', 0.01) * self.cfg.get('sample_rate', 16000)
        )  # 160 samples for every 1ms by default
        self.num_mel_frame_per_target_frame = int(self.cfg.get('subsampling_factor', 8))

    def __getitem__(self, cuts) -> Tuple[torch.Tensor, ...]:
        # NOTE: This end-to-end diarization dataloader only loads the 1st ch of the audio file.
        # Process cuts in a single loop: convert to mono and compute speaker activities
        mono_cuts = []
        speaker_activities = []
        for cut in cuts:
            if cut.num_channels is not None and cut.num_channels > 1:
                logging.warning(
                    "Multiple channels detected in cut '%s' (%d channels). "
                    "Only the first channel will be used; remaining channels are ignored.",
                    cut.id,
                    cut.num_channels,
                )
            mono_cut = cut.with_channels(channels=[0])
            mono_cuts.append(mono_cut)

            speaker_activity = speaker_to_target(
                a_cut=mono_cut,
                num_speakers=self.num_speakers,
                num_sample_per_mel_frame=self.num_sample_per_mel_frame,
                num_mel_frame_per_asr_frame=self.num_mel_frame_per_target_frame,
                boundary_segments=True,
            )
            # This line prevents dimension mismatch error in the collate_matrices function.
            if speaker_activity.shape[1] > self.num_speakers:
                logging.warning(
                    "Number of speakers in the target %s is greater than "
                    "the maximum number of speakers %s. Truncating extra speakers. "
                    "Set the `num_speakers` to higher value to avoid this warning.",
                    speaker_activity.shape[1],
                    self.num_speakers,
                )
                speaker_activity = speaker_activity[:, : self.num_speakers]
            speaker_activities.append(speaker_activity)

        cuts = type(cuts).from_cuts(mono_cuts)
        audio, audio_lens, cuts = self.load_audio(cuts)
        targets = collate_matrices(speaker_activities).to(audio.dtype)  # (B, T, N)

        if targets.shape[2] > self.num_speakers:
            targets = targets[:, :, : self.num_speakers]
        elif targets.shape[2] < self.num_speakers:
            targets = torch.nn.functional.pad(
                targets, (0, self.num_speakers - targets.shape[2]), mode='constant', value=0
            )

        target_lens_list = []
        for audio_len in audio_lens:
            target_fr_len = get_hidden_length_from_sample_length(
                audio_len, self.num_sample_per_mel_frame, self.num_mel_frame_per_target_frame
            )
            target_lens_list.append(target_fr_len)
        target_lens = torch.tensor(target_lens_list)

        return audio, audio_lens, targets, target_lens



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from nemo.collections.asr.data.feature_to_text import FeatureToBPEDataset, FeatureToCharDataset
from nemo.utils import logging


def get_char_dataset(config: dict, augmentor: Optional['FeatureAugmentor'] = None) -> FeatureToCharDataset:
    """
    Instantiates a Character Encoding based FeatureToCharDataset.

    Args:
        config: Config of the FeatureToCharDataset.
        augmentor: Optional AudioAugmentor object for augmentations on audio data.

    Returns:
        An instance of FeatureToCharDataset.
    """
    if 'labels' not in config:
        logging.warning(f"dataset does not have explicitly defined labels")

    dataset = FeatureToCharDataset(
        manifest_filepath=config['manifest_filepath'],
        labels=config.get('labels', None),
        normalize=config.get('normalize', 'post_norm'),
        normalize_type=config.get('normalize_type', 'per_feature'),
        use_rttm=config.get('use_rttm', False),
        rttm_mode=config.get('rttm_mode', 'mask'),
        feat_min_len=config.get('feat_min_len', 4),
        feat_mask_val=config.get('feat_mask_val', None),
        frame_unit_time_secs=config.get('frame_unit_time_secs', 0.01),
        sample_rate=config.get('sample_rate', 16000),
        augmentor=augmentor,
        max_duration=config.get('max_duration', None),
        min_duration=config.get('min_duration', None),
        max_utts=config.get('max_utts', 0),
        blank_index=config.get('blank_index', -1),
        unk_index=config.get('unk_index', -1),
        trim=config.get('trim_silence', False),
        parser=config.get('parser', 'en'),
        return_sample_id=config.get('return_sample_id', False),
        channel_selector=config.get('channel_selector', None),
    )
    return dataset


def get_bpe_dataset(
    config: dict, tokenizer: 'TokenizerSpec', augmentor: Optional['FeatureAugmentor'] = None
) -> FeatureToBPEDataset:
    """
    Instantiates a Byte Pair Encoding / Word Piece Encoding based FeatureoToBPEDataset.

    Args:
        config: Config of the FeatureToBPEDataset.
        tokenizer: An instance of a TokenizerSpec object.
        augmentor: Optional FeatureAugmentor object for augmentations on audio features.

    Returns:
        An instance of FeatureToBPEDataset.
    """
    dataset = FeatureToBPEDataset(
        manifest_filepath=config['manifest_filepath'],
        tokenizer=tokenizer,
        normalize=config.get('normalize', 'post_norm'),
        normalize_type=config.get('normalize_type', 'per_feature'),
        use_rttm=config.get('use_rttm', False),
        rttm_mode=config.get('rttm_mode', 'mask'),
        feat_min_len=config.get('feat_min_len', 4),
        feat_mask_val=config.get('feat_mask_val', None),
        frame_unit_time_secs=config.get('frame_unit_time_secs', 0.01),
        sample_rate=config.get('sample_rate', 16000),
        augmentor=augmentor,
        max_duration=config.get('max_duration', None),
        min_duration=config.get('min_duration', None),
        max_utts=config.get('max_utts', 0),
        trim=config.get('trim_silence', False),
        use_start_end_token=config.get('use_start_end_token', True),
        return_sample_id=config.get('return_sample_id', False),
        channel_selector=config.get('channel_selector', None),
    )
    return dataset



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import lightning.pytorch as pl
import pytest
import torch
from omegaconf import OmegaConf

from nemo.core.classes import ModelPT
from nemo.utils.exp_manager import exp_manager

try:
    # `ptl_resiliency` is included in `gwe_resiliency_pkg` package
    from ptl_resiliency import StragglerDetectionCallback

    HAVE_STRAGGLER_DET = True
except (ImportError, ModuleNotFoundError):
    HAVE_STRAGGLER_DET = False


class OnesDataset(torch.utils.data.Dataset):
    def __init__(self, dataset_len):
        super().__init__()
        self.__dataset_len = dataset_len

    def __getitem__(self, *args):
        return torch.ones(2)

    def __len__(self):
        return self.__dataset_len


class ExampleModel(ModelPT):
    def __init__(self, log_dir, **kwargs):
        cfg = OmegaConf.structured({})
        super().__init__(cfg)
        pl.seed_everything(1234)
        self.l1 = torch.nn.modules.Linear(in_features=2, out_features=1)
        self.log_dir = log_dir

    def on_train_start(self):
        super().on_train_start()
        rank = torch.distributed.get_rank()

    def train_dataloader(self):
        dataset = OnesDataset(1024 * 1024)
        return torch.utils.data.DataLoader(dataset, batch_size=2, num_workers=2)

    def val_dataloader(self):
        dataset = OnesDataset(128 * 1024)
        return torch.utils.data.DataLoader(dataset, batch_size=2, num_workers=2)

    def forward(self, batch):
        output = self.l1(batch)
        output = torch.nn.functional.l1_loss(output, torch.zeros(output.size()).to(output.device))
        return output

    def validation_step(self, batch, batch_idx):
        self.loss = self(batch)
        return self.loss

    def training_step(self, batch, batch_idx):
        return self(batch)

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters(), lr=0.1)

    def list_available_models(self, *args, **kwargs):
        pass

    def setup_training_data(self, *args, **kwargs):
        pass

    def setup_validation_data(self, *args, **kwargs):
        pass

    def on_validation_epoch_end(self):
        self.log("val_loss", torch.stack([self.loss]).mean())


@pytest.mark.skipif(not HAVE_STRAGGLER_DET, reason="requires resiliency package to be installed.")
class TestStragglerDetection:

    @pytest.mark.run_only_on('GPU')
    def test_prints_perf_scores(self, tmp_path):
        # Run dummy 1 rank DDP training
        # Training time is limited to 3 seconds and straggler reporting is set to 1 second
        # Check if there are straggler related logs in the captured log
        max_steps = 1_000_000
        tmp_path = tmp_path / "test_1"
        print("TMP PATH", tmp_path)

        trainer = pl.Trainer(
            strategy='ddp',
            devices=1,
            accelerator='gpu',
            enable_checkpointing=False,
            logger=False,
            max_steps=max_steps,
            val_check_interval=0.33,
        )
        exp_manager(
            trainer,
            {
                "max_time_per_run": "00:00:00:03",
                "explicit_log_dir": str(tmp_path),
                "create_checkpoint_callback": False,
                "create_straggler_detection_callback": True,
                "straggler_detection_params": {
                    "report_time_interval": 1.0,
                    "calc_relative_gpu_perf": True,
                    "calc_individual_gpu_perf": True,
                    "num_gpu_perf_scores_to_log": 1,
                },
            },
        )
        model = ExampleModel(log_dir=tmp_path)
        trainer.fit(model)

        # assume that NeMo logs are written into "nemo_log_globalrank-0_localrank-0.txt"
        rank0_log_content = None
        with open(tmp_path / "nemo_log_globalrank-0_localrank-0.txt") as f:
            rank0_log_content = f.read()

        assert "GPU relative performance" in rank0_log_content
        assert "GPU individual performance" in rank0_log_content



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional

import torch

from nemo.core.classes import NeuralModule, typecheck
from nemo.core.neural_types import NeuralType, SpectrogramType


class MixtureConsistencyProjection(NeuralModule):
    """Ensure estimated sources are consistent with the input mixture.
    Note that the input mixture is assume to be a single-channel signal.

    Args:
        weighting: Optional weighting mode for the consistency constraint.
            If `None`, use uniform weighting. If `power`, use the power of the
            estimated source as the weight.
        eps: Small positive value for regularization

    Reference:
        Wisdom et al, Differentiable consistency constraints for improved deep speech enhancement, 2018
    """

    def __init__(self, weighting: Optional[str] = None, eps: float = 1e-8):
        super().__init__()
        self.weighting = weighting
        self.eps = eps

        if self.weighting not in [None, 'power']:
            raise NotImplementedError(f'Weighting mode {self.weighting} not implemented')

    @property
    def input_types(self) -> Dict[str, NeuralType]:
        """Returns definitions of module output ports."""
        return {
            "mixture": NeuralType(('B', 'C', 'D', 'T'), SpectrogramType()),
            "estimate": NeuralType(('B', 'C', 'D', 'T'), SpectrogramType()),
        }

    @property
    def output_types(self) -> Dict[str, NeuralType]:
        """Returns definitions of module output ports."""
        return {
            "output": NeuralType(('B', 'C', 'D', 'T'), SpectrogramType()),
        }

    @typecheck()
    def forward(self, mixture: torch.Tensor, estimate: torch.Tensor) -> torch.Tensor:
        """Enforce mixture consistency on the estimated sources.
        Args:
            mixture: Single-channel mixture, shape (B, 1, F, N)
            estimate: M estimated sources, shape (B, M, F, N)

        Returns:
            Source estimates consistent with the mixture, shape (B, M, F, N)
        """
        if mixture.size(-3) != 1:
            raise ValueError(f'Mixture must have a single channel, got shape {mixture.shape}')

        # number of sources
        M = estimate.size(-3)
        # estimated mixture based on the estimated sources
        estimated_mixture = torch.sum(estimate, dim=-3, keepdim=True)

        # weighting
        if self.weighting is None:
            weight = 1 / M
        elif self.weighting == 'power':
            weight = estimate.abs().pow(2)
            weight = weight / (weight.sum(dim=-3, keepdim=True) + self.eps)
        else:
            raise NotImplementedError(f'Weighting mode {self.weighting} not implemented')

        # consistent estimate
        consistent_estimate = estimate + weight * (mixture - estimated_mixture)

        return consistent_estimate



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional

import lightning.pytorch as pl
import nemo_run as run
import torch
from lightning.pytorch.callbacks.callback import Callback
from megatron.core.distributed import DistributedDataParallelConfig
from megatron.core.optimizer import OptimizerConfig

from nemo import lightning as nl
from nemo.collections.diffusion.data.diffusion_mock_datamodule import MockDataModule
from nemo.collections.diffusion.models.flux.model import FluxModelParams, MegatronFluxModel
from nemo.collections.llm.api import pretrain
from nemo.collections.llm.recipes.log.default import default_log, tensorboard_logger
from nemo.collections.llm.recipes.precision.mixed_precision import bf16_mixed
from nemo.utils.exp_manager import TimingCallback

NAME = "flux_12b"


@run.cli.factory
@run.autoconvert
def flux_mock_datamodule() -> pl.LightningDataModule:
    """Mock Datamodule Initialization"""
    data_module = MockDataModule(
        image_h=1024,
        image_w=1024,
        micro_batch_size=1,
        global_batch_size=2,
        image_precached=True,
        text_precached=True,
    )
    return data_module


@run.cli.factory(name=NAME)
def model() -> run.Config[pl.LightningModule]:
    """
    Factory function to create a FLUX 12B model configuration.

    Returns:
        run.Config[pl.LightningModule]: Configuration for the FLUX 12B model.

    """
    return run.Config(MegatronFluxModel, flux_params=run.Config(FluxModelParams))


def trainer(
    tensor_parallelism: int = 1,
    pipeline_parallelism: int = 1,
    pipeline_parallelism_type: Optional[torch.dtype] = None,
    virtual_pipeline_parallelism: Optional[int] = None,
    context_parallelism: int = 1,
    sequence_parallelism: bool = False,
    num_nodes: int = 1,
    num_gpus_per_node: int = 8,
    max_steps: int = 1168251,
    callbacks: Optional[list[run.Config[Callback]]] = None,
) -> run.Config[nl.Trainer]:
    """
    Configure the NeMo Lightning Trainer for FLUX 12B model.

    This function sets up the distributed training strategy and other training parameters.

    Args:
        tensor_parallelism (int): Degree of tensor model parallelism.
        pipeline_parallelism (int): Degree of pipeline model parallelism.
        pipeline_parallelism_type (Optional[torch.dtype]): Data type for pipeline parallelism.
        virtual_pipeline_parallelism (Optional[int]): Size of virtual pipeline parallelism.
        context_parallelism (int): Degree of context parallelism.
        sequence_parallelism (bool): Whether to use sequence parallelism.
        num_nodes (int): Number of compute nodes to use.
        num_gpus_per_node (int): Number of GPUs per node.
        max_steps (int): Maximum number of training steps.
        callbacks (Optional[list[run.Config[Callback]]]): List of callback configurations.

    Returns:
        run.Config[nl.Trainer]: Configuration for the NeMo Lightning Trainer.

    Examples:
        CLI usage:
            $ nemo llm pretrain trainer=flux_12b ...

        Python API usage:
            >>> trainer_config = trainer(num_nodes=2, num_gpus_per_node=8)
            >>> print(trainer_config)

    Note:
        For more information on distributed training strategies, refer to the
        NeMo documentation on multi-GPU and multi-node training.
    """
    strategy = run.Config(
        nl.MegatronStrategy,
        tensor_model_parallel_size=tensor_parallelism,
        pipeline_model_parallel_size=pipeline_parallelism,
        context_parallel_size=context_parallelism,
        sequence_parallel=sequence_parallelism,
        pipeline_dtype=pipeline_parallelism_type,
        gradient_accumulation_fusion=True,
        ddp=run.Config(
            DistributedDataParallelConfig,
            # data_parallel_sharding_strategy='optim_grads_params',
            check_for_nan_in_grad=True,
            grad_reduce_in_fp32=True,
            # overlap_grad_reduce=True,
            # overlap_param_gather=True,
        ),
    )

    trainer = run.Config(
        nl.Trainer,
        accelerator="gpu",
        accumulate_grad_batches=1,
        callbacks=callbacks,
        devices=num_gpus_per_node,
        limit_test_batches=50,
        limit_val_batches=32,
        log_every_n_steps=10,
        max_steps=max_steps,
        num_nodes=num_nodes,
        plugins=bf16_mixed(),
        strategy=strategy,
        use_distributed_sampler=False,
        val_check_interval=2000,
    )

    return trainer


@run.cli.factory(target=pretrain, name=NAME)
def pretrain_recipe(
    dir: Optional[str] = None,
    name: str = "default",
    num_nodes: int = 1,
    num_gpus_per_node: int = 8,
    fn: Callable = pretrain,
) -> run.Partial:
    """
    Create a pre-training recipe for FLUX 12B model.

    This function sets up a complete configuration for pre-training, including
    model, trainer, data, logging, optimization, and resumption settings.

    Args:
        dir (Optional[str]): Directory for saving logs and checkpoints.
        name (str): Name of the pre-training run.
        num_nodes (int): Number of compute nodes to use.
        num_gpus_per_node (int): Number of GPUs per node.
        fn (Callable): The pre-training function to use.

    Returns:
        run.Partial: Partial configuration for pre-training.

    Examples:
        CLI usage:
            $ nemo llm pretrain --factory flux_12b

        Python API usage:
            >>> recipe = pretrain_recipe(name="flux_12b_pretrain", num_nodes=1)
            >>> print(recipe)

    Note:
        For more details on pre-training LLMs with NeMo, see the pre-training
        guide in the `examples/llm/pretrain/` directory.
    """
    recipe = run.Partial(
        fn,
        model=model(),
        trainer=trainer(
            num_nodes=num_nodes,
            num_gpus_per_node=num_gpus_per_node,
            callbacks=[run.Config(TimingCallback)],
        ),
        data=flux_mock_datamodule(),
        log=default_log(dir=dir, name=name, tensorboard_logger=tensorboard_logger(name=name)),
        optim=run.Config(
            nl.MegatronOptimizerModule,
            config=run.Config(
                OptimizerConfig,
                lr=1e-4,
                bf16=True,
                use_distributed_optimizer=True,
                weight_decay=0,
            ),
        ),
        resume=run.Config(
            nl.AutoResume,
            resume_if_exists=True,
            resume_ignore_no_checkpoint=True,
            resume_past_end=True,
        ),
        # model_transform=None,
    )

    return recipe



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from functools import partial
from pathlib import Path
from typing import Any, Callable, Dict, Optional

import lightning.pytorch as pl
from lightning.fabric.utilities.types import _PATH
from lightning.pytorch.trainer.states import TrainerFn
from lightning.pytorch.utilities import model_summary
from typing_extensions import override

from nemo.collections.llm.fn import base as fn
from nemo.lightning.ckpt_utils import ADAPTER_META_FILENAME
from nemo.lightning.io.pl import ckpt_to_dir
from nemo.lightning.megatron_parallel import MegatronParallel
from nemo.lightning.pytorch.callbacks.peft import PEFT, WrappedAdapterIO
from nemo.utils import logging

SPEECHLM_PEFT_RESUME = 'speechlm_peft_resume'


class SpeechToTextLLMPEFT(PEFT):
    def __init__(self, peft: PEFT):
        super().__init__()
        self.peft = peft

    def get_wrappped_io(self):
        """
        This is a helper function to return a partial function that wraps the checkpoint I/O with the PEFT adapter.
        """
        return partial(SpeechLMWrappedAdapterIO, peft=self)

    def __call__(
        self, model: "nemo.collections.speechlm.model.SpeechToTextLLM"  # noqa: F821
    ) -> "nemo.collections.speechlm.model.SpeechToTextLLM":  # noqa: F821
        """Apply the PEFT method to the LLM.

        This method freezes the LLM parameters and walks through the model
        structure, applying the transform method to LLM module.

        Args:
            model (nn.Module): The model to be fine-tuned.

        Returns:
            nn.Module: The transformed model with PEFT applied.
        """
        # only apply PEFT to the language model
        model.freeze_llm()
        module = model.module
        logging.info(f"Applying PEFT to language model with: {self.peft}")

        # If using megatron virtual pipeline parallelism, model is a list of
        # model chunks so iterate over model
        if isinstance(model, MegatronParallel) and len(model) > 1:
            for model_chunk in model:
                self._transform_module(model_chunk)
        else:
            self._transform_module(module)

        if hasattr(model, "trainer") and model.trainer.state.fn != TrainerFn.FITTING:
            self.freeze_model(model)

        logging.info(f"\n{model_summary.summarize(model, max_depth=4)}")
        return model

    def _transform_module(self, module):
        while not hasattr(module, "language_model"):
            module = module.module
        fn.walk(module.language_model, self.transform, _skip_map=True)

    def transform(self, module, name=None, prefix=None):
        return self.peft.transform(module, name=name, prefix=prefix)

    def set_params_to_save(self, trainer: pl.Trainer):
        """
        Set params that should be saved for PEFT, including some params that don't require gradients,
        such as the running mean and var of batchnorm.
        """
        model = trainer.lightning_module  # type: nemo.collections.speechlm.model.SpeechToTextLLM # noqa: F821
        self.params_to_save = set([name for name, _ in model.trainable_parameters()])
        if len(self.params_to_save) == 0:
            raise RuntimeError("No trainable parameters found for PEFT!")


class SpeechLMWrappedAdapterIO(WrappedAdapterIO):

    @override
    def load_checkpoint(
        self,
        path: _PATH,
        sharded_state_dict=None,
        map_location: Optional[Callable] = None,
        strict: Optional['StrictHandling'] | bool = None,
    ) -> Dict[str, Any]:
        """
        Overwrite the load_checkpoint method to handle PEFT resume for SpeechLM.

        =====================
        Initial PEFT Training
        =====================
        Initial PEFT training requires loading the base model weights. In this case, this function is called by
        trainer.strategy.setup() -> megatron_strategy.restore_model() -> megatron_strategy.load_checkpoint().
        `path = PosixPath(<base_path>)`, and sharded_state_dict contains only base model weights

        ===========
        PEFT Resume
        ===========
        PEFT resume requires loading two set of model weights, 1) base model weights and 2) adapter weights
        Base model weights could be imported from e.g. HF, and is frozen during PEFT training.
        Adapter weights contains the training metadata that will need to be loaded.
        As such, this function will be entered twice during PEFT training resume.

        For the FIRST TIME this function is called by trainer._checkpoint_connector._restore_modules_and_callbacks.
        `path = AdapterPath(<adapter_path>, base_model_path=<base_path>)`, and sharded_state_dict contains only base model weights

        For the SECOND TIME this function is called by PEFT.apply_transform (above, in the same file).
        `path = PosixPath(<adapter_path>)`, and sharded_state_dict contains only adapter weights.
        """

        assert self.checkpoint_io is not None

        adapter_meta_path = ckpt_to_dir(path) / ADAPTER_META_FILENAME
        adapter_ckpt = None
        load_base = False

        if getattr(path, "base_model_path", None):
            ## PEFT Resume, FIRST TIME
            self.adapter_ckpt_path = Path(str(path))
            adapter_ckpt = self.checkpoint_io.load_checkpoint(path, sharded_state_dict={})  # Loads only metadata
            # path is adapter path to restore the training metadata, but switch to loading base model here.
            path = self.model_ckpt_path = path.base_model_path
            load_base = True
        elif adapter_meta_path.exists():
            ## PEFT Resume, SECOND TIME
            with open(adapter_meta_path, "r") as f:
                metadata = json.load(f)
            self.model_ckpt_path = Path(metadata['model_ckpt_path'])
            self.adapter_ckpt_path = path
        else:
            ## Initial PEFT Training
            self.model_ckpt_path = path

        # Note: this will include the Trainer-state of the model-checkpoint
        model_ckpt = self._load_checkpoint(path, sharded_state_dict, map_location, load_base, strict=strict)

        if adapter_ckpt is not None:
            ## PEFT Resume, FIRST TIME
            adapter_ckpt['state_dict'].update(model_ckpt['state_dict'])
            if SPEECHLM_PEFT_RESUME in model_ckpt:
                adapter_ckpt[SPEECHLM_PEFT_RESUME] = True
            return adapter_ckpt
        return model_ckpt

    def _load_checkpoint(
        self,
        path: _PATH,
        sharded_state_dict,
        map_location: Optional[Callable] = None,
        load_base: bool = False,
        strict: Optional['StrictHandling'] | bool = None,
    ) -> None:
        if load_base:
            # Return empty state_dict to skip loading PTL checkpoint in first stage of PEFT
            # Must use with nemo.collections.speechlm.strategies.megatron_strategy.SpeechLMMegatronStrategy
            return {'state_dict': dict(), SPEECHLM_PEFT_RESUME: True}
        else:
            model_ckpt = self.checkpoint_io.load_checkpoint(path, sharded_state_dict, map_location, strict)

        return model_ckpt



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from torch import nn

from nemo.collections.multimodal.speech_cv.parts.submodules.conv2d import Conv2d
from nemo.collections.multimodal.speech_cv.parts.submodules.global_avg_pool2d import GlobalAvgPool2d
from nemo.collections.multimodal.speech_cv.parts.submodules.resnet_block import ResNetBlock
from nemo.collections.multimodal.speech_cv.parts.submodules.resnet_bottleneck_block import ResNetBottleneckBlock


class ResNet(nn.Module):

    """ ResNet (ResNet18, ResNet34, ResNet50, ResNet101, ResNet152)
    Models: 224 x 224
    ResNet18: 11,689,512 Params
    ResNet34: 21,797,672 Params
    ResNet50: 25,557,032 Params
    ResNet101: 44,549,160 Params
    Resnet152: 60,192,808 Params
    Reference: "Deep Residual Learning for Image Recognition" by He et al.
    https://arxiv.org/abs/1512.03385
    """

    def __init__(self, dim_input=3, dim_output=1000, model="ResNet50", include_stem=True, include_head=True):
        super(ResNet, self).__init__()

        assert model in ["ResNet18", "ResNet34", "ResNet50", "ResNet101", "ResNet152"]

        if model == "ResNet18":
            dim_stem = 64
            dim_blocks = [64, 128, 256, 512]
            num_blocks = [2, 2, 2, 2]
            bottleneck = False
        elif model == "ResNet34":
            dim_stem = 64
            dim_blocks = [64, 128, 256, 512]
            num_blocks = [3, 4, 6, 3]
            bottleneck = False
        elif model == "ResNet50":
            dim_stem = 64
            dim_blocks = [256, 512, 1024, 2048]
            num_blocks = [3, 4, 6, 3]
            bottleneck = True
        elif model == "ResNet101":
            dim_stem = 64
            dim_blocks = [256, 512, 1024, 2048]
            num_blocks = [3, 4, 23, 3]
            bottleneck = True
        elif model == "ResNet152":
            dim_stem = 64
            dim_blocks = [256, 512, 1024, 2048]
            num_blocks = [3, 8, 36, 3]
            bottleneck = True

        self.stem = (
            nn.Sequential(
                Conv2d(
                    in_channels=dim_input,
                    out_channels=dim_stem,
                    kernel_size=(7, 7),
                    stride=(2, 2),
                    weight_init="he_normal",
                    bias=False,
                ),
                nn.BatchNorm2d(num_features=dim_stem),
                nn.ReLU(),
                nn.MaxPool2d(kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
            )
            if include_stem
            else nn.Identity()
        )

        # Blocks
        self.blocks = nn.ModuleList()
        for stage_id in range(4):

            for block_id in range(num_blocks[stage_id]):

                # Projection Block
                if block_id == 0:
                    if stage_id == 0:
                        stride = (1, 1)
                        bottleneck_ratio = 1
                        in_features = dim_stem
                    else:
                        stride = (2, 2)
                        bottleneck_ratio = 2
                        in_features = dim_blocks[stage_id - 1]
                # Default Block
                else:
                    stride = (1, 1)
                    in_features = dim_blocks[stage_id]
                    bottleneck_ratio = 4

                if bottleneck:
                    self.blocks.append(
                        ResNetBottleneckBlock(
                            in_features=in_features,
                            out_features=dim_blocks[stage_id],
                            bottleneck_ratio=bottleneck_ratio,
                            kernel_size=(3, 3),
                            stride=stride,
                        )
                    )
                else:
                    self.blocks.append(
                        ResNetBlock(
                            in_features=in_features,
                            out_features=dim_blocks[stage_id],
                            kernel_size=(3, 3),
                            stride=stride,
                        )
                    )

        # Head
        self.head = (
            nn.Sequential(
                GlobalAvgPool2d(),
                nn.Linear(in_features=dim_blocks[-1], out_features=dim_output)
                if dim_output is not None
                else nn.Identity(),
            )
            if include_head
            else nn.Identity()
        )

    def forward(self, x):

        # Is Video
        if x.dim() == 5:

            is_video = True
            batch_size = x.shape[0]
            video_frames = x.shape[2]

            # (B, Din, T, H, W) -> (B * T, Din, H, W)
            x = x.transpose(1, 2).flatten(start_dim=0, end_dim=1)

        else:
            is_video = False

        # (B, Din, H, W) -> (B, D0, H//4, W//4)
        x = self.stem(x)

        # (B, D0, H//4, W//4) -> (B, D4, H//32, W//32)
        for block in self.blocks:
            x = block(x)

        # (B, D4, H//32, W//32) -> (B, Dout)
        x = self.head(x)

        # Is Video
        if is_video:

            # (B * T, Dout) -> (B, Dout, T)
            if x.dim() == 2:
                x = x.reshape(batch_size, video_frames, -1).transpose(1, 2)

            # (B * T, D4, H//32, W//32) -> (B, D4, T, H//32, W//32)
            else:
                x = x.reshape(batch_size, video_frames, x.shape[1], x.shape[2], x.shape[3]).transpose(1, 2)

        return x



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
from unittest.mock import MagicMock

from lightning.pytorch.callbacks import Callback as PTLCallback

from nemo.lightning.base_callback import BaseCallback


def _fresh_group_module():
    """Reset the CallbackGroup singleton and stub OneLoggerNeMoCallback safely.

    This avoids deleting modules from sys.modules. We import the module,
    replace the OneLoggerNeMoCallback symbol with a lightweight stub,
    and reset the internal singleton so a new instance is built.
    """
    mod = importlib.import_module('nemo.lightning.callback_group')

    class _StubOneLoggerCallback(BaseCallback):
        def __init__(self, *args, **kwargs):
            pass

        def update_config(self, *args, **kwargs):
            pass

    setattr(mod, 'OneLoggerNeMoCallback', _StubOneLoggerCallback)
    # Reset the singleton so the next get_instance() uses the stubbed class
    mod.CallbackGroup._instance = None
    return mod


def test_base_callback_noops_do_not_raise():
    """Test BaseCallback hooks are no-ops and do not raise exceptions."""
    cb = BaseCallback()

    cb.on_app_start()
    cb.on_app_end()
    cb.on_model_init_start()
    cb.on_model_init_end()
    cb.on_dataloader_init_start()
    cb.on_dataloader_init_end()
    cb.on_optimizer_init_start()
    cb.on_optimizer_init_end()
    cb.on_load_checkpoint_start()
    cb.on_load_checkpoint_end()
    cb.on_save_checkpoint_start()
    cb.on_save_checkpoint_end()
    cb.on_save_checkpoint_success()
    cb.update_config()


def test_base_callback_is_ptl_callback():
    """Test BaseCallback derives from Lightning PTL Callback."""
    assert isinstance(BaseCallback(), PTLCallback)


def test_callback_group_singleton_identity():
    """Test CallbackGroup returns the same singleton instance."""
    mod = _fresh_group_module()
    a = mod.CallbackGroup.get_instance()
    b = mod.CallbackGroup.get_instance()
    assert a is b


def test_callback_group_update_config_fanout_and_attach(monkeypatch):
    """Test update_config fans out to callbacks and attaches them to trainer."""
    mod = _fresh_group_module()
    group = mod.CallbackGroup.get_instance()

    class _StubCallback(BaseCallback):
        def __init__(self):
            self.called = False
            self.kwargs = None

        def update_config(self, *args, **kwargs):
            self.called = True
            self.kwargs = kwargs

    stub_cb = _StubCallback()
    group._callbacks = [stub_cb]

    class Trainer:
        def __init__(self):
            self.callbacks = []

    trainer = Trainer()
    marker = object()
    group.update_config('v2', trainer, data=marker)

    assert stub_cb.called
    kwargs = stub_cb.kwargs
    assert kwargs['nemo_version'] == 'v2'
    assert kwargs['trainer'] is trainer
    assert kwargs['data'] is marker
    assert trainer.callbacks[0] is stub_cb


def test_callback_group_dynamic_dispatch_calls_when_present():
    """Test dynamic dispatch calls methods when present on callbacks."""
    mod = _fresh_group_module()
    group = mod.CallbackGroup.get_instance()

    mock_cb = MagicMock()
    group._callbacks = [mock_cb]

    group.on_app_start()
    assert mock_cb.on_app_start.called


def test_callback_group_dynamic_dispatch_ignores_missing_methods():
    """Test dynamic dispatch ignores missing methods without raising."""
    mod = _fresh_group_module()
    group = mod.CallbackGroup.get_instance()

    class Dummy:
        pass

    group._callbacks = [Dummy()]

    # Should not raise even if method not present
    group.on_nonexistent_method()


def test_hook_class_init_with_callbacks_wraps_and_emits(monkeypatch):
    """Test inheritance-based hook via __init_subclass__ emits start/end once (e2e-style)."""
    mod = _fresh_group_module()
    group = mod.CallbackGroup.get_instance()

    start = MagicMock()
    end = MagicMock()

    monkeypatch.setattr(group, 'on_model_init_start', start)
    monkeypatch.setattr(group, 'on_model_init_end', end)

    class Base:
        def __init_subclass__(cls, **kwargs):
            super().__init_subclass__(**kwargs)
            # Mirror IOMixin: hook subclasses at definition time
            mod.hook_class_init_with_callbacks(cls, 'on_model_init_start', 'on_model_init_end')

    class Child(Base):
        def __init__(self):
            self.x = 1

    class GrandChild(Child):
        def __init__(self):
            self.y = 2
            super().__init__()

    c = Child()
    assert c.x == 1
    # Flag indicating wrapping applied on the subclass
    assert getattr(Child.__init__, '_init_wrapped_for_callbacks', False) is True

    d = GrandChild()
    assert d.x == 1
    assert d.y == 2

    assert start.call_count == 2
    assert end.call_count == 2
    # Flag indicating wrapping applied on the subclass
    assert getattr(GrandChild.__init__, '_init_wrapped_for_callbacks', False) is True


def test_hook_class_init_with_callbacks_idempotent():
    """Test inheritance-based hook is idempotent and does not re-wrap on repeated calls."""
    mod = _fresh_group_module()

    class Base:
        def __init_subclass__(cls, **kwargs):
            super().__init_subclass__(**kwargs)
            mod.hook_class_init_with_callbacks(cls, 'on_model_init_start', 'on_model_init_end')

    class Child(Base):
        def __init__(self):
            pass

    # Hook was applied via __init_subclass__ at class creation time
    first = Child.__init__
    # Attempt to apply again explicitly; should be a no-op
    mod.hook_class_init_with_callbacks(Child, 'on_model_init_start', 'on_model_init_end')
    second = Child.__init__
    assert first is second


def test_on_app_end_is_idempotent(monkeypatch):
    """Test on_app_end fans out only once even if called multiple times."""
    mod = _fresh_group_module()
    group = mod.CallbackGroup.get_instance()

    mock_cb = MagicMock()
    group._callbacks = [mock_cb]

    group.on_app_end()
    group.on_app_end()

    assert mock_cb.on_app_end.call_count == 1



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest.mock import patch

import torch
from transformers import AutoProcessor

from nemo.collections.multimodal.data.energon.config import ImageToken, LLaVATemplateConfig, MultiModalSampleConfig
from nemo.collections.vlm.llava_next.data.sample import (
    LlavaNextTextRawBatch,
    LlavaNextTextSample,
    PackedLlavaNextTextSample,
)
from nemo.collections.vlm.llava_next.data.task_encoder import LlavaNextTaskEncoder
from nemo.collections.vlm.llava_next.data.vqa_sample_encoder import LlavaNextSampleEncoder


class TestLlavaNextTaskEncoder(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Use actual processor
        cls.processor = AutoProcessor.from_pretrained("llava-hf/llava-v1.6-vicuna-7b-hf")
        cls.tokenizer = cls.processor.tokenizer
        cls.image_processor = cls.processor.image_processor

    def setUp(self):
        self.config = MultiModalSampleConfig(
            image_token=ImageToken(token_str="<image>", token_id=-200),
            ignore_place_holder=-100,
            conversation_template_config=LLaVATemplateConfig(
                system="I am an AI assistant", roles=["user", "assistant"], stop_string=" </s>"
            ),
        )
        self.encoder = LlavaNextTaskEncoder(
            tokenizer=self.tokenizer, image_processor=self.image_processor, multimodal_sample_config=self.config
        )

    def test_batch(self):
        # Import VQA sample types
        from megatron.energon import VQASample

        # Create VQA sample encoder
        vqa_encoder = LlavaNextSampleEncoder(
            tokenizer=self.tokenizer, image_processor=self.image_processor, multimodal_sample_config=self.config
        )

        # Create various VQA samples with different context/answer lengths and image sizes
        vqa_samples = [
            # Sample 1: Short context, short answer, small image
            VQASample(
                __key__="sample1",
                context="What is in this image?",
                answers="A cat.",
                image=torch.rand(3, 224, 224),  # Small image
                __restore_key__=None,
                __subflavor__=None,
                __subflavors__=[],
            ),
            # Sample 2: Medium context, medium answer, medium image
            VQASample(
                __key__="sample2",
                context="Can you describe what's happening in this image and identify the main objects?",
                answers="The image shows a living room with a sofa, coffee table, and two people sitting and talking.",
                image=torch.rand(3, 336, 336),  # Medium image
                __restore_key__=None,
                __subflavor__=None,
                __subflavors__=[],
            ),
            # Sample 3: Long context, long answer, large image
            VQASample(
                __key__="sample3",
                context="This is a complex scene with multiple elements. Can you analyze what's happening, describe the setting, identify any people or objects present?",
                answers="The image depicts an outdoor cafe scene on a busy street. There are approximately 8-10 people sitting at various tables, enjoying their meals and conversations. The setting appears to be in a European city, given the architecture visible in the background.",
                image=torch.rand(3, 448, 448),  # Large image
                __restore_key__=None,
                __subflavor__=None,
                __subflavors__=[],
            ),
        ]

        # Encode each sample
        encoded_samples = []
        for sample in vqa_samples:
            output_sample = LlavaNextTextSample()
            encoded_sample = vqa_encoder.encode(sample, output_sample)
            encoded_samples.append(encoded_sample)

        # Create batch using the encoder
        batch = self.encoder.batch(encoded_samples)

        # Print shapes of key tensors in batch
        print(f"Batch tokens shape: {batch.tokens.shape}")  # [num_samples, max_seq_len]
        print(f"Batch labels shape: {batch.labels.shape}")  # [num_samples, max_seq_len]
        print(f"Batch images shape: {batch.images.shape}")  # [total_tiles, channels, height, width]
        print(f"Batch num_media_tiles: {batch.num_media_tiles}")  # [num_samples]

        # Verify batch properties
        self.assertIsInstance(batch, LlavaNextTextRawBatch)
        self.assertEqual(len(batch.__keys__), len(encoded_samples))
        self.assertEqual(batch.tokens.shape[0], len(encoded_samples))
        self.assertEqual(batch.tokens.shape[1], max(sample.tokens.shape[0] for sample in encoded_samples))
        self.assertEqual(batch.labels.shape[0], len(encoded_samples))
        self.assertEqual(batch.labels.shape[1], max(sample.labels.shape[0] for sample in encoded_samples))
        self.assertEqual(batch.images.shape[0], sum(batch.num_media_tiles).item())
        self.assertEqual(len(batch.num_media_tiles), len(encoded_samples))

    @patch('nemo.collections.vlm.neva.data.sequence_packing.greedy_knapsack')
    @patch('nemo.collections.vlm.neva.data.sequence_packing.predict_seq_len_with_padding')
    def test_select_samples_to_pack(self, mock_predict_seq_len_with_padding, mock_greedy_knapsack):
        # Configure encoder with sequence packing enabled
        packing_encoder = LlavaNextTaskEncoder(
            tokenizer=self.tokenizer,
            image_processor=self.image_processor,
            multimodal_sample_config=self.config,
            packed_sequence=True,
            packed_sequence_size=1024,
        )

        # Create test samples with different token lengths and image sizes
        samples = [
            # Sample 1: Short sequence, small image
            LlavaNextTextSample(
                __key__="sample1",
                tokens=torch.tensor([1, 2, 3, 4, 5], dtype=torch.long),
                labels=torch.tensor([1, 2, 3, 4, 5], dtype=torch.long),
                images=torch.rand(1, 3, 224, 224),  # 1 tile, small image
                loss_mask=torch.tensor([1.0, 1.0, 1.0, 1.0, 1.0]),
                num_media_tiles=1,
                image_sizes=torch.tensor([[224, 224]], dtype=torch.long),
                attention_mask=torch.tensor([1, 1, 1, 1, 1], dtype=torch.long),
            ),
            # Sample 2: Medium sequence, medium image
            LlavaNextTextSample(
                __key__="sample2",
                tokens=torch.tensor([4, 5, 6, 7, 8, 9, 10, 11, 12], dtype=torch.long),
                labels=torch.tensor([4, 5, 6, 7, 8, 9, 10, 11, 12], dtype=torch.long),
                images=torch.rand(2, 3, 336, 336),  # 2 tiles, medium image
                loss_mask=torch.tensor([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]),
                num_media_tiles=2,
                image_sizes=torch.tensor([[336, 336]], dtype=torch.long),
                attention_mask=torch.tensor([1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=torch.long),
            ),
            # Sample 3: Long sequence, large image
            LlavaNextTextSample(
                __key__="sample3",
                tokens=torch.tensor([13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25], dtype=torch.long),
                labels=torch.tensor([13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25], dtype=torch.long),
                images=torch.rand(4, 3, 448, 448),  # 4 tiles, large image
                loss_mask=torch.tensor([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]),
                num_media_tiles=4,
                image_sizes=torch.tensor([[448, 448]], dtype=torch.long),
                attention_mask=torch.tensor([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=torch.long),
            ),
        ]

        # Configure mocks
        mock_predict_seq_len_with_padding.side_effect = [5, 9, 13]  # Return token lengths
        mock_greedy_knapsack.return_value = [[samples[0], samples[1]], [samples[2]]]  # Return packed groups

        # Call the method
        result = packing_encoder.select_samples_to_pack(samples)

        # Verify the results
        mock_greedy_knapsack.assert_called_once()
        self.assertEqual(len(result), 2)  # Two bins created
        self.assertEqual(len(result[0]), 2)  # Two samples in first bin
        self.assertEqual(len(result[1]), 1)  # One sample in second bin
        self.assertEqual(result[0][0].__key__, "sample1")
        self.assertEqual(result[0][1].__key__, "sample2")
        self.assertEqual(result[1][0].__key__, "sample3")

    @patch('nemo.collections.vlm.llava_next.data.utils.convert_to_packed_llava_next')
    def test_pack_selected_samples(self, mock_convert_to_packed):
        # Configure encoder with sequence packing enabled
        packing_encoder = LlavaNextTaskEncoder(
            tokenizer=self.tokenizer,
            image_processor=self.image_processor,
            multimodal_sample_config=self.config,
            packed_sequence=True,
            packed_sequence_size=1024,
        )

        # Create samples to pack with different token lengths and image sizes
        samples = [
            # Sample 1: Short sequence, small image
            LlavaNextTextSample(
                __key__="sample1",
                tokens=torch.tensor([1, 2, 3, 4, 5], dtype=torch.long),
                labels=torch.tensor([1, 2, 3, 4, 5], dtype=torch.long),
                images=torch.rand(1, 3, 336, 336),  # 1 tile, small image
                loss_mask=torch.tensor([1.0, 1.0, 1.0, 1.0, 1.0]),
                num_media_tiles=1,
                image_sizes=torch.tensor([[336, 336]], dtype=torch.long),
                attention_mask=torch.tensor([1, 1, 1, 1, 1], dtype=torch.long),
            ),
            # Sample 2: Medium sequence, medium image
            LlavaNextTextSample(
                __key__="sample2",
                tokens=torch.tensor([6, 7, 8, 9, 10, 11, 12], dtype=torch.long),
                labels=torch.tensor([6, 7, 8, 9, 10, 11, 12], dtype=torch.long),
                images=torch.rand(2, 3, 336, 336),  # 2 tiles, medium image
                loss_mask=torch.tensor([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]),
                num_media_tiles=2,
                image_sizes=torch.tensor([[336, 336]], dtype=torch.long),
                attention_mask=torch.tensor([1, 1, 1, 1, 1, 1, 1], dtype=torch.long),
            ),
        ]

        # Configure mock for convert_to_packed_llava_next
        # Returned values need to match the combined lengths of the samples above
        mock_convert_to_packed.return_value = (
            torch.tensor([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]),  # packed_tokens
            torch.tensor([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]),  # packed_labels
            torch.tensor([0, 1, 2, 3, 4, 0, 1, 2, 3, 4, 5, 6]),  # packed_position_ids
            torch.tensor([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]),  # packed_loss_mask
            {
                "sample_lengths": [5, 7],
                "sequence_index_map": torch.tensor([0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1]),
            },  # packed_seq_params
        )

        # Call the method
        result = packing_encoder.pack_selected_samples(samples)

        # Verify the result
        self.assertIsInstance(result, PackedLlavaNextTextSample)
        self.assertEqual(result.__key__, "sample1,sample2")
        self.assertEqual(result.images.shape[0], 3)  # 1+2 tiles

        # Update assertions to match actual tensor shapes
        # The tokens are padded to length 128 and batched with dim 1
        self.assertEqual(result.tokens.shape, torch.Size([1, 128]))  # Padded from original 12 tokens to 128
        self.assertEqual(result.labels.shape, torch.Size([1, 128]))  # Similarly padded
        self.assertEqual(result.loss_mask.shape, torch.Size([1, 128]))  # Similarly padded

        # Check for packed_seq_params fields
        # cu_seqlens_q/kv [0,5,12] represents cumulative sequence lengths:
        # - 0: Start index of first sequence
        # - 5: End of first sequence (length 5) / Start of second sequence
        # - 12: End of second sequence (length 7)
        self.assertIsNotNone(result.packed_seq_params)
        self.assertEqual(result.packed_seq_params.qkv_format, 'thd')
        self.assertEqual(len(result.packed_seq_params.cu_seqlens_q), 3)  # [0, 5, 12]
        self.assertEqual(result.packed_seq_params.cu_seqlens_q[0].item(), 0)  # Start index
        self.assertEqual(result.packed_seq_params.cu_seqlens_q[1].item(), 5)  # After first sequence
        self.assertEqual(result.packed_seq_params.cu_seqlens_q[2].item(), 12)  # After second sequence

        # Check num_media_tiles
        self.assertEqual(result.num_media_tiles.shape, torch.Size([2]))  # 2 images worth of tiles
        self.assertEqual(result.image_sizes.shape[0], 2)  # 2 images worth of sizes


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
from torch import Tensor

from nemo.collections.asr.inference.model_wrappers.asr_inference_wrapper import ASRInferenceWrapper
from nemo.collections.asr.models import EncDecHybridRNNTCTCModel, EncDecRNNTModel


class RNNTInferenceWrapper(ASRInferenceWrapper):
    """
    Provides a unified interface to work with RNNT/TDT/Hybrid models.
    """

    def __post_init__(self) -> None:
        """
        Additional post initialization step
        Checks if the model is a rnnt model and sets the decoding strategy to rnnt.
        """
        if not isinstance(self.asr_model, (EncDecRNNTModel, EncDecHybridRNNTCTCModel)):
            raise ValueError(
                "Provided model is not a RNNT type. You are trying to use a RNNT transcriber with a non-RNNT model."
            )

        decoder_type = 'rnnt'
        if isinstance(self.asr_model, EncDecHybridRNNTCTCModel):
            self.asr_model.cur_decoder = decoder_type

        # reset the decoding strategy
        self.reset_decoding_strategy(decoder_type)
        self.set_decoding_strategy(decoder_type)

        self.cast_dtype = torch.float32 if self.use_amp else self.compute_dtype
        self.asr_model.to(self.cast_dtype)

    def get_blank_id(self) -> int:
        """
        Returns id of the blank token.
        Returns:
            (int) blank id for the model.
        """
        blank_id = len(self.asr_model.joint.vocabulary)
        return blank_id

    def get_vocabulary(self) -> list[str]:
        """
        Returns the list of vocabulary tokens.
        Returns:
            (list[str]) list of vocabulary tokens.
        """
        return self.asr_model.joint.vocabulary

    def get_subsampling_factor(self) -> int:
        """
        Returns the subsampling factor for the ASR encoder.
        Returns:
            (int) subsampling factor for the ASR encoder model.
        """
        return self.asr_model.encoder.subsampling_factor

    def encode(
        self, processed_signal: Tensor, processed_signal_length: Tensor, prompt_vectors: Tensor | None = None
    ) -> tuple[Tensor, Tensor]:
        """
        Get encoder output from the model. It is used for streaming inference.
        Args:
            processed_signal: (Tensor) processed signal. Shape is torch.Size([B, C, T]).
            processed_signal_length: (Tensor) processed signal length. Shape is torch.Size([B]).
            prompt_vectors: (Tensor | None) Optional prompt vectors for multilingual models.
                Shape can be torch.Size([B, num_prompts]) or torch.Size([B, T_enc, num_prompts]) if already expanded.
        Returns:
            (tuple[Tensor, Tensor]) encoder output and encoder output length of shape torch.Size([B, T, D]), torch.Size([B]).
        """
        if processed_signal.device != self.device:
            processed_signal = processed_signal.to(self.device)

        if processed_signal_length.device != self.device:
            processed_signal_length = processed_signal_length.to(self.device)

        with (
            torch.amp.autocast(device_type=self.device_str, dtype=self.compute_dtype, enabled=self.use_amp),
            torch.inference_mode(),
            torch.no_grad(),
        ):

            # Prepare model arguments
            model_args = {
                'processed_signal': processed_signal.to(self.cast_dtype),
                'processed_signal_length': processed_signal_length,
            }
            if prompt_vectors is not None:
                model_args['prompt'] = prompt_vectors

            forward_outs = self.asr_model(**model_args)

        encoded, encoded_len = forward_outs
        return encoded, encoded_len

    def decode(self, encoded: Tensor, encoded_len: Tensor, partial_hypotheses: list) -> list:
        """
        Decode the encoder output using the RNNT decoder.
        Args:
            encoded: (Tensor) encoder output.
            encoded_len: (Tensor) encoder output length.
            partial_hypotheses: (list) list of partial hypotheses for stateful decoding.
        Returns:
            (list) list of best hypotheses.
        """
        best_hyp = self.asr_model.decoding.rnnt_decoder_predictions_tensor(
            encoded.to(self.cast_dtype), encoded_len, return_hypotheses=True, partial_hypotheses=partial_hypotheses
        )
        return best_hyp

    def encode_with_prompts(
        self, processed_signal: Tensor, processed_signal_length: Tensor, prompt_vectors: Tensor
    ) -> tuple[Tensor, Tensor]:
        """
        Convenience wrapper for prompt-enabled encoding.
        Expands prompt vectors across the time dimension before calling encode.
        Args:
            processed_signal: (Tensor) processed signal. Shape is torch.Size([B, C, T]).
            processed_signal_length: (Tensor) processed signal length. Shape is torch.Size([B]).
            prompt_vectors: (Tensor) prompt vectors. Shape is torch.Size([B, num_prompts]).
        Returns:
            (tuple[Tensor, Tensor]) encoder output and encoder output length.
        """
        encoder_time_steps = processed_signal.shape[2] // self.get_subsampling_factor()
        # Expand prompts: [B, num_prompts] -> [B, T_enc, num_prompts]
        prompt_vectors = prompt_vectors.unsqueeze(1).expand(-1, encoder_time_steps, -1)
        return self.encode(
            processed_signal=processed_signal,
            processed_signal_length=processed_signal_length,
            prompt_vectors=prompt_vectors,
        )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import logging as _logging
import os
import re
import shutil
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Dict, Literal, Optional, Union

import lightning.pytorch as pl
import torch
import torch.distributed as dist
from lightning.fabric.plugins import CheckpointIO
from lightning.fabric.utilities.rank_zero import rank_zero_info
from lightning.fabric.utilities.seed import reset_seed
from lightning.pytorch.strategies.model_parallel import ModelParallelStrategy as PLModelParallelStrategy
from lightning.pytorch.trainer.states import TrainerFn
from lightning.pytorch.utilities.types import STEP_OUTPUT
from torch.distributed.tensor.parallel import ColwiseParallel, RowwiseParallel, SequenceParallel
from typing_extensions import override

from nemo.lightning import io
from nemo.lightning.pytorch.strategies.utils import (
    _destroy_dist_connection,
    ckpt_to_dir,
    create_checkpoint_io,
    fsdp2_strategy_parallelize,
)
from nemo.utils import logging
from nemo.utils.import_utils import safe_import_from

try:
    from torch.distributed.tensor._api import distribute_tensor
    from torch.distributed.tensor.placement_types import Replicate, Shard
except ImportError:
    from torch.distributed._tensor.api import distribute_tensor
    from torch.distributed._tensor.placement_types import Replicate, Shard

MixedPrecisionPolicy, HAS_MIXED_PRECISION_POLICY = safe_import_from(
    "torch.distributed.fsdp", "MixedPrecisionPolicy", fallback_module="torch.distributed._composable.fsdp"
)
CPUOffloadPolicy, HAS_CPU_OFFLOAD_POLICY = safe_import_from(
    "torch.distributed.fsdp", "CPUOffloadPolicy", fallback_module="torch.distributed._composable.fsdp"
)

_logger = _logging.getLogger(__name__)


class FSDP2Strategy(PLModelParallelStrategy, io.IOMixin):
    """FSDP2Strategy implementing FSDP via FSDP 2."""

    def __init__(
        self,
        data_parallel_size: Union[Literal["auto"], int] = "auto",
        tensor_parallel_size: Union[Literal["auto"], int] = "auto",
        context_parallel_size: Optional[int] = 1,
        sequence_parallel: bool = False,
        offload_policy: 'CPUOffloadPolicy' = None,
        data_sampler=None,
        checkpoint_io=None,
        mp_policy=None,
        parallelize_fn=fsdp2_strategy_parallelize,
        use_hf_tp_plan: bool = True,
        custom_tp_plan: Optional[Dict[str, Union[RowwiseParallel, ColwiseParallel, SequenceParallel]]] = None,
        **kwargs,
    ):
        """Initializes the FSDP2Strategy with specified parallelization settings.

        Args:
            data_parallel_size (Union[Literal["auto"], int]): Size of data parallel. Defaults to "auto".
            tensor_parallel_size (Union[Literal["auto"], int]): Size of tensor parallel. Defaults to "auto".
            context_parallel_size (optional): Number of context-parallel groups. Defaults to 1.
            sequence_parallel (bool): Whether to enable sequence parallelism when use_hf_tp_plan is False and
                custom_tp_plan is not provided. Defaults to False. Only effective when tensor_parallel_size > 1.
            data_sampler (optional): Custom data sampler to process dataloaders.
            mp_policy (optional): Mixed precision policy for parameter and operation casting.
                Defaults to:
                ```python
                MixedPrecisionPolicy(
                    param_dtype=torch.bfloat16,
                    reduce_dtype=torch.float32,
                    output_dtype=None,
                    cast_forward_inputs=True,
                )
                ```
            parallelize_fn (callable, optional): Function for parallelizing the model. Defaults to None.
            use_hf_tp_plan (bool, optional): Whether to use the huggingface TP plan. This will be used if
                custom_tp_plan is not provided. Also, sequence_parallel option will be ignored if use_hf_tp_plan
                is set to True. Defaults to True.
            custom_tp_plan (Optional[Dict[str, Any]], optional): Custom tensor parallel plan for the model.
                tensor_parallel_size need to be > 1 to use this option. If provided, it overrides the
                default tensor parallel plan. sequence_parallel option will be ignored if custom_tp_plan
                is provided.
            **kwargs: Additional arguments for base class initialization.
        """
        super().__init__(data_parallel_size=data_parallel_size, tensor_parallel_size=tensor_parallel_size, **kwargs)
        self._checkpoint_io = checkpoint_io
        self.context_parallel_size = context_parallel_size
        self.data_sampler = data_sampler
        self.checkpoint = None
        self.mp_policy = mp_policy
        if self.mp_policy is None:
            assert HAS_MIXED_PRECISION_POLICY is not None, "Expected to have MixedPrecisionPolicy"
            self.mp_policy = MixedPrecisionPolicy(
                param_dtype=torch.bfloat16,
                reduce_dtype=torch.bfloat16,
                output_dtype=torch.bfloat16,
                cast_forward_inputs=True,
            )
        self.store: Optional[torch.distributed.Store] = None
        self.parallelize_fn = parallelize_fn
        self.offload_policy = offload_policy
        self.sequence_parallel = sequence_parallel
        self.use_hf_tp_plan = use_hf_tp_plan

        self.tp_shard_plan = None
        if custom_tp_plan is not None:
            self.tp_shard_plan = custom_tp_plan
            logging.info(
                "You are using a custom TP plan. Make sure it is compatible with the model. Parallelization would ",
                "not raise errors if the custom TP plan is not compatible. SP option will also be ignored.",
            )
        elif self.use_hf_tp_plan:
            logging.info(
                "You are using a huggingface TP plan. Make sure your model is a huggingface model. Certain ",
                "parallelizations might not be supported. SP option will also be ignored.",
            )
        else:
            # Parallelize the first embedding and the last linear out projection
            base_model_tp_plan = {
                "model.embed_tokens": RowwiseParallel(input_layouts=Replicate()),
                "model.layers.*.self_attn.q_proj": ColwiseParallel(),
                "model.layers.*.self_attn.k_proj": ColwiseParallel(),
                "model.layers.*.self_attn.v_proj": ColwiseParallel(),
                "model.layers.*.self_attn.o_proj": RowwiseParallel(),
                "model.layers.*.mlp.up_proj": ColwiseParallel(),
                "model.layers.*.mlp.gate_proj": ColwiseParallel(),
                "model.layers.*.mlp.down_proj": RowwiseParallel(),
                "lm_head": ColwiseParallel(output_layouts=Replicate()),
            }

            base_model_sp_plan = {
                "model.embed_tokens": RowwiseParallel(input_layouts=Replicate(), output_layouts=Shard(1)),
                "model.norm": SequenceParallel(),
                "model.layers.*.input_layernorm": SequenceParallel(),
                "model.layers.*.self_attn.o_proj": RowwiseParallel(output_layouts=Shard(1)),
                "model.layers.*.post_attention_layernorm": SequenceParallel(),
                "model.layers.*.mlp.down_proj": RowwiseParallel(output_layouts=Shard(1)),
                "lm_head": ColwiseParallel(input_layouts=Shard(1), output_layouts=Replicate()),
            }

            if self.sequence_parallel:
                # Enable sequence parallelism only if TP size > 1
                base_model_tp_plan.update(base_model_sp_plan)

            self.tp_shard_plan = base_model_tp_plan
            logging.info(
                "Using default TP plan for parallelization. It is compatible with huggingface llama3-style models."
            )

    @property
    @override
    def lightning_restore_optimizer(self) -> bool:
        """Optim state restoration is enabled"""
        return True

    def load_optimizer_state_dict(self, checkpoint) -> None:
        """Stores a reference to the optimizer state-dict for later restoration.

        Instead of immediately restoring the optimizer's state, this method saves the checkpoint
        reference and defers the restoration until the first training step. This is necessary
        because, in NeMo 2.0, PeFT adapters are added dynamically just before the first training
        step. Attempting to restore the optimizer state-dict before the adapters are initialized
        would result in an error.

        Args:
            checkpoint (dict): A dictionary containing the trainer's checkpoint,
                            including the optimizer state-dict.
        """
        # TODO(@akoumparouli): refactor.
        self.checkpoint = checkpoint

    def _load_optimizer_state_dict(self) -> None:
        """Restores the optimizer state-dict from the stored checkpoint.

        This method applies the optimizer state stored in the checkpoint to the corresponding
        optimizers. It ensures that the optimizer states are correctly restored after the
        PeFT adapters have been added in the first training step.

        If no checkpoint is stored, the method exits without performing any restoration.

        Note: This operation runs only once, as the checkpoint reference is cleared after execution.
        """
        from torch.distributed.checkpoint.state_dict import set_optimizer_state_dict

        if self.checkpoint is None:
            for optimizer, opt_state in zip(self.optimizers, self.checkpoint["optimizer_states"]):
                set_optimizer_state_dict(
                    self.lightning_module,
                    optimizer,
                    optim_state_dict=opt_state,
                    options={},
                )
            # run this only once
            self.checkpoint = None

    @override
    def setup_environment(self) -> None:
        """setup distributed environment and device mesh"""
        from torch.distributed.device_mesh import init_device_mesh

        self.accelerator.setup_device(self.root_device)

        self._setup_distributed()
        if self._data_parallel_size == "auto":
            self._data_parallel_size = self.num_nodes

        if self._tensor_parallel_size == "auto":
            self._tensor_parallel_size = self.num_processes

        mesh_shape = []
        mesh_dim_names = []
        # TP needs to be the last dimension as innermost dimension, DP-CP-TP
        for dim, name in zip(
            [self._data_parallel_size, self.context_parallel_size, self._tensor_parallel_size],
            ["data_parallel", "context_parallel", "tensor_parallel"],
        ):
            mesh_shape.append(int(dim))
            mesh_dim_names.append(name)

        self._device_mesh = init_device_mesh(
            device_type=self.root_device.type,
            mesh_shape=tuple(mesh_shape),
            mesh_dim_names=mesh_dim_names,
        )

        # Construct sharding and reduction meshes for specific configurations.
        # Replace existing mesh strategies if a custom mesh design is provided.
        # WARNING: ADDING MESHES INCREASES MEMORY USAGE. Use device meshes for
        # multiple dimensions of parallelism if possible.
        if self._device_mesh["context_parallel"].size() > 1:
            # Context parallelism loss reduction mesh. Remember to divide out CP size in tokens per second throughput.
            self._device_mesh[("data_parallel", "context_parallel")]._flatten(mesh_dim_name="dp_cp")

        self.lightning_module._device_mesh = self._device_mesh

    @override
    def setup(self, trainer: pl.Trainer) -> None:
        """Configures the strategy within the PyTorch Lightning trainer.

        Args:
            trainer (pl.Trainer): The PyTorch Lightning trainer instance.
        """
        self.trainer = trainer
        # connect trainer to accelerator.
        self.accelerator.setup(trainer)
        # Parallelize model
        if getattr(self, '_init_model_parallel', True):
            self.parallelize()
        # Corner case, as FSDP2 expected to be used multi-device.
        if self._data_parallel_size == 1:
            self._lightning_module = self._lightning_module.to(self.root_device)
        # setup optim
        if getattr(self, '_setup_optimizers', True) and trainer.state.fn == TrainerFn.FITTING:
            super().setup_optimizers(trainer)

    def parallelize(self):
        """Applies fully_shard on model"""
        if self.parallelize_fn is not None:
            # TODO(@akoumparouli): self.lightning_module is an nn.Module child, use it directly?
            # Apply FSDP2 and TP to the model
            self.parallelize_fn(
                self.lightning_module.model,
                device_mesh=self._device_mesh,
                mp_policy=self.mp_policy,
                use_hf_tp_plan=self.use_hf_tp_plan,
                tp_shard_plan=self.tp_shard_plan,
                offload_policy=self.offload_policy,
            )
            # Apply this only once
            self.parallelize_fn = None
        else:
            logging.warning("Called parallelize more than once.")

    @override
    def _setup_distributed(self) -> None:
        """Initializes process group for communications."""

        # Implementation from superclass copied below in order to pass the store to the process group init
        reset_seed()
        self.set_world_ranks()
        self._process_group_backend = self._get_process_group_backend()
        # See https://github.com/pytorch/pytorch/issues/148532 for details.
        if self.offload_policy is not None:
            self._process_group_backend = "cuda:nccl,cpu:gloo"

        assert self.cluster_environment is not None
        if not torch.distributed.is_available():
            raise RuntimeError("torch.distributed is not available. Cannot initialize distributed process group")
        if torch.distributed.is_initialized():
            _logger.debug("torch.distributed is already initialized. Exiting early")
            return

        global_rank = self.cluster_environment.global_rank()
        world_size = self.cluster_environment.world_size()
        os.environ["MASTER_ADDR"] = self.cluster_environment.main_address
        os.environ["MASTER_PORT"] = str(self.cluster_environment.main_port)
        _logger.info(f"Initializing distributed: GLOBAL_RANK: {global_rank}, MEMBER: {global_rank + 1}/{world_size}")
        torch.distributed.init_process_group(
            self._process_group_backend, rank=global_rank, world_size=world_size, store=self.store
        )

        if self._process_group_backend == "nccl":
            atexit.register(_destroy_dist_connection)

        # On rank=0 let everyone know training is starting
        rank_zero_info(
            f"{'-' * 100}\n"
            f"distributed_backend={self._process_group_backend}\n"
            f"All distributed processes registered. Starting with {world_size} processes\n"
            f"{'-' * 100}\n"
        )

    def _get_loss_reduction(self, step_type: str):
        """Retrieves the loss reduction method for a given step type.

        Args:
            step_type (str): The type of step (e.g., "training", "validation").

        Returns:
            Callable: The loss reduction function, if defined; otherwise, None.
        """
        for fn_name in [f"{step_type}_loss_reduction", "loss_reduction"]:
            if hasattr(self.lightning_module, fn_name):
                return getattr(self.lightning_module, fn_name)
        return None

    def _step_proxy(self, step_type, batch, batch_idx=None):
        """Executes a training, validation, or test step and applies loss reduction if available.

        Args:
            step_type (str): The step type ("training", "validation", "test", "predict").
            batch: The input batch.
            batch_idx (optional): Index of the batch.

        Returns:
            Tuple: The computed loss and a dictionary with reduced loss metrics.
        """
        method_name = f"{step_type}_step"
        if self.model != self.lightning_module:
            loss = self._forward_redirection(self.model, self.lightning_module, method_name, batch, batch_idx)
        else:
            loss = getattr(self.lightning_module, method_name)(batch, batch_idx)

        _loss_reduction = self._get_loss_reduction(step_type)
        if _loss_reduction:
            return _loss_reduction.forward(batch, loss)
        return loss, {'avg': loss}

    @override
    def optimizer_state(self, optimizer):
        """returns the sharded optim state"""
        return optimizer.state_dict()

    @override
    def training_step(self, batch, batch_idx=None) -> STEP_OUTPUT:
        """Defines the training step, logging relevant metrics.

        Args:
            batch: The input batch.
            batch_idx (optional): The index of the batch.

        Returns:
            STEP_OUTPUT: The loss for backpropagation.
        """

        # See load_optimizer_state_dict to understand why we call this here.
        if self.checkpoint is not None:
            self._load_optimizer_state_dict()

        assert self.lightning_module is not None
        assert self.model is not None

        if self.context_parallel_size > 1:
            # Only pass context_parallel=True if AutoModel supports and has non-trivial CP.
            loss = self.lightning_module.training_step(batch, batch_idx, context_parallel=True)
        else:
            loss = self.lightning_module.training_step(batch, batch_idx)

        self.lightning_module.log(
            'global_step',
            self.trainer.global_step,
            prog_bar=True,
            rank_zero_only=True,
            batch_size=1,
        )

        self.lightning_module.log(
            'step',
            self.trainer.global_step,
        )

        return loss

    @override
    def validation_step(self, batch, batch_idx=None) -> Any:
        """Defines the validation step, logging validation loss.

        Args:
            batch: The input batch.
            batch_idx (optional): The index of the batch.

        Returns:
            Any: The validation loss.
        """
        assert self.lightning_module is not None
        assert self.model is not None
        with self.precision_plugin.val_step_context():
            loss, reduced = self._step_proxy("validation", batch, batch_idx)
            if reduced["avg"]:
                self.lightning_module.log('val_loss', reduced['avg'], rank_zero_only=True, batch_size=1)
            return loss

    @override
    def test_step(self, batch, batch_idx=None) -> STEP_OUTPUT:
        """Defines the test step, logging test loss.

        Args:
            batch: The input batch.
            batch_idx (optional): The index of the batch.

        Returns:
            Any: The test loss.
        """
        assert self.lightning_module is not None
        assert self.model is not None
        with self.precision_plugin.test_step_context():
            loss, reduced = self._step_proxy("test", batch, batch_idx)
            self.lightning_module.log('test_loss', reduced['avg'], rank_zero_only=True, batch_size=1)

            return loss

    @override
    def predict_step(self, batch, batch_idx=None) -> STEP_OUTPUT:
        """Runs one predict step.

        Args:
            batch (dict): the batch to use for pred.
            batch_idx (int, optional): the batch index. Defaults to None.

        Returns:
            STEP_OUTPUT: the reduced loss.
        """
        assert self.lightning_module is not None
        assert self.model is not None
        with self.precision_plugin.predict_step_context():
            loss, reduced = self._step_proxy("predict", batch, batch_idx)
            return reduced

    @contextmanager
    @override
    def tensor_init_context(self, empty_init: Optional[bool] = None):
        """Context manager used for initialization"""
        # Materializaton happens in `setup()`
        # @akoumparouli: using Parent's tensor_init_context causes mcore
        # parameters to be initialized on GPU instead of (assumed) CPU.
        yield

    @property
    @override
    def checkpoint_io(self) -> CheckpointIO:
        """CheckpointIO getter

        Returns:
            CheckpointIO: _description_
        """
        if self._checkpoint_io is None:
            self._checkpoint_io = create_checkpoint_io()

        return self._checkpoint_io

    @checkpoint_io.setter
    def checkpoint_io(self, io: CheckpointIO) -> None:
        """CheckpointIO setter

        Args:
            io (CheckpointIO): the checkpointio to use.
        """
        self._checkpoint_io = io

    @property
    def current_epoch_step(self) -> int:
        """Gets the current step within an epoch.

        Returns:
            int: The step index within the epoch.
        """
        return max(
            self.trainer.fit_loop.epoch_loop.automatic_optimization.optim_progress.optimizer.step.current.completed,
            self.trainer.fit_loop.epoch_loop.manual_optimization.optim_step_progress.current.completed,
        )

    @override
    def remove_checkpoint(self, filepath: Union[str, Path]) -> None:
        """Removes a checkpoint from the filesystem.

        Args:
            filepath (Union[str, Path]): Path to the checkpoint to be removed.
        """
        ckpt = ckpt_to_dir(filepath)
        if self.is_global_zero:
            if os.path.islink(ckpt):
                os.unlink(ckpt)
            elif os.path.exists(ckpt):
                shutil.rmtree(ckpt)

    @override
    def lightning_module_state_dict(self) -> Dict[str, Any]:
        """Collects the state dict of the model."""
        from nemo.lightning.pytorch.strategies.utils import to_cpu

        assert self.lightning_module is not None
        tmp_sd = self.lightning_module.state_dict()
        # Get adapter_only value from checkpoint io. There's two cases:
        # - nemo.lightning.pytorch.callbacks.peft.WrappedAdapterIO
        #   In this case, the self._checkpoint_io object is a wrapper and holds a `checkpoint_io`
        #   attribute which we query for the `adapter_only` attribute
        # - otherwise, it's the base case which has the adapter_only attribute directly accesible.
        is_adapter_only = getattr(
            self._checkpoint_io,
            'adapter_only',
            getattr(getattr(self._checkpoint_io, 'checkpoint_io', {}), 'adapter_only', False),
        )

        if is_adapter_only:
            # if any key has "lora" in FQN, then it will only move lora keys to cpu, since only
            # the adapter weights are saved.
            name_has_lora = lambda x: 'lora' in x.lower()
            module_names = list(filter(name_has_lora, tmp_sd.keys()))
        else:
            module_names = list(tmp_sd.keys())

        state_dict = {}
        for name in module_names:
            param = tmp_sd.pop(name)
            state_dict[name] = to_cpu(param)

        dist.barrier()
        return state_dict

    @override
    def save_checkpoint(
        self, checkpoint: Dict[str, Any], filepath: Union[str, Path], storage_options: Optional[Any] = None
    ) -> None:
        """
        Unshards FSDP2 checkpoint and passes it to checkpoint_io for saving to a file.
        """

        if self.is_global_zero:
            self.checkpoint_io.save_checkpoint(checkpoint, filepath, storage_options=storage_options)

    @override
    def load_checkpoint(self, checkpoint_path: str | Path) -> Dict[str, Any]:
        """Loads checkpoint with checkpoint_io"""
        return self.checkpoint_io.load_checkpoint(checkpoint_path)

    @override
    @torch.no_grad
    def load_model_state_dict(
        self,
        ckpt,
        strict=False,
    ):
        """Shards a full state dict"""
        if self._tensor_parallel_size > 1 and self._device_mesh["context_parallel"].size() == 1:
            # Gather TP/SP strategy keys
            colwise_keys = [k for k in self.tp_shard_plan if isinstance(self.tp_shard_plan[k], ColwiseParallel)]
            rowwise_keys = [k for k in self.tp_shard_plan if isinstance(self.tp_shard_plan[k], RowwiseParallel)]
            seq_parallel_keys = [k for k in self.tp_shard_plan if isinstance(self.tp_shard_plan[k], SequenceParallel)]

            sharded_state = {k: v for k, v in ckpt['state_dict'].items()}

            # placement is (dp, tp)
            for k, v in sharded_state.items():
                if any(re.match(x, k) for x in seq_parallel_keys):
                    sharded_state[k] = distribute_tensor(v, self.device_mesh, placements=(Shard(dim=0), Replicate()))
                elif any(re.match(x, k) for x in colwise_keys):
                    sharded_state[k] = distribute_tensor(v, self.device_mesh, placements=(Shard(dim=0), Shard(dim=0)))
                elif any(re.match(x, k) for x in rowwise_keys):
                    sharded_state[k] = distribute_tensor(v, self.device_mesh, placements=(Shard(dim=0), Shard(dim=1)))
                else:
                    # This is for layers not sharded by TP/SP
                    sharded_state[k] = distribute_tensor(
                        v, self.device_mesh["data_parallel"], placements=(Shard(dim=0),)
                    )
        # TODO(@akoumparouli): update `placements` value once TP is enabled.
        elif self._tensor_parallel_size == 1 and self._device_mesh["context_parallel"].size() > 1:
            # Shard across the CP device mesh, associated with the fully_shard() call
            # in utils.fsdp2_strategy_parallelize().
            sharded_state = {
                k: distribute_tensor(
                    v, self._device_mesh[("data_parallel", "context_parallel")], placements=(Replicate(), Shard(dim=0))
                )
                for k, v in ckpt['state_dict'].items()
            }
        else:
            # Default shard across DP for FSDP2.
            sharded_state = {
                k: distribute_tensor(v, self._device_mesh["data_parallel"], placements=(Shard(dim=0),))
                for k, v in ckpt['state_dict'].items()
            }

        self.lightning_module.load_state_dict(sharded_state, strict=strict)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import logging
import os
import shutil
from collections import OrderedDict
from pathlib import Path
from typing import Any, Dict, Optional, Union

import lightning.pytorch as pl
import torch
from lightning.fabric.plugins import CheckpointIO
from lightning.fabric.strategies.fsdp import _get_sharded_state_dict_context
from lightning.fabric.utilities.rank_zero import rank_zero_info
from lightning.fabric.utilities.seed import reset_seed
from lightning.pytorch.strategies.fsdp import FSDPStrategy as PLFSDPStrategy
from lightning.pytorch.trainer.states import TrainerFn
from lightning.pytorch.utilities.types import STEP_OUTPUT

try:
    from megatron.core.transformer.transformer_layer import TransformerLayer

    HAVE_MEGATRON_CORE = True

except (ImportError, ModuleNotFoundError):

    TransformerLayer = object
    HAVE_MEGATRON_CORE = False

from torch.distributed.checkpoint.state_dict import (  # get_state_dict,
    StateDictOptions,
    get_optimizer_state_dict,
    set_state_dict,
)
from torch.utils.data import DataLoader
from typing_extensions import override

from nemo.lightning import io
from nemo.lightning.pytorch.strategies.utils import (
    _destroy_dist_connection,
    ckpt_to_dir,
    create_checkpoint_io,
    fix_progress_bar,
    init_model_parallel,
    mcore_to_pyt_sharded_state_dict,
    pyt_to_mcore_state_dict,
    setup_data_sampler,
    setup_parallel_ranks,
)

_logger = logging.getLogger(__name__)


class FSDPStrategy(PLFSDPStrategy, io.IOMixin):
    """Megatron plugin for Pytorch Lightning.

    This strategy implements Fully-Sharded-Data-Parallel using PyTorch's native FSDP methods.
    Comparing with MegatronStrategy, FSDPStrategy is designed to be more lightweight, with
    minimal modifications over Lightning's FSDPStrategy but preserves necessary features to be
    compatible with nemo and mcore.
    By default, this strategy wraps FSDP per TransformerLayer.

    Note:
        This strategy is designed to work with NVIDIA's Megatron-LM framework and requires
        specific model implementations that are compatible with Megatron's parallelism techniques.
    Note:
        Due to the different optimizer structure (FSDP only uses torch native optimizers),
        MegatronStrategy cannot resume training from checkpoints saved by FSDPStrategy, and vice
        versa. However, the model weights structure is made compatible, so switching strategy is
        possible if users only need the weights not the optimizer states. (E.g. run pretrain with
        megatron 4D parallelism and run SFT with FSDP.)
    """

    def __init__(
        self,
        auto_wrap_policy={TransformerLayer},
        state_dict_type="sharded",
        ckpt_load_optimizer: bool = True,
        ckpt_save_optimizer: bool = True,
        data_sampler=None,
        **kwargs,
    ):
        super().__init__(auto_wrap_policy=auto_wrap_policy, state_dict_type=state_dict_type, **kwargs)

        self.data_sampler = data_sampler
        self.ckpt_load_optimizer = ckpt_load_optimizer
        self.ckpt_save_optimizer = ckpt_save_optimizer
        self.store: Optional[torch.distributed.Store] = None

    @override
    def setup_environment(self) -> None:
        """Initializes rank and process group for communications."""
        setup_parallel_ranks(self)

        self.accelerator.setup_device(self.root_device)

        # Implementation from superclass copied below in order to pass the store to the process group init
        reset_seed()
        self.set_world_ranks()
        self._process_group_backend = self._get_process_group_backend()
        assert self.cluster_environment is not None

        if not torch.distributed.is_available():
            raise RuntimeError("torch.distributed is not available. Cannot initialize distributed process group")
        if torch.distributed.is_initialized():
            _logger.debug("torch.distributed is already initialized. Exiting early")
            return

        global_rank = self.cluster_environment.global_rank()
        world_size = self.cluster_environment.world_size()
        os.environ["MASTER_ADDR"] = self.cluster_environment.main_address
        os.environ["MASTER_PORT"] = str(self.cluster_environment.main_port)
        _logger.info(f"Initializing distributed: GLOBAL_RANK: {global_rank}, MEMBER: {global_rank + 1}/{world_size}")
        torch.distributed.init_process_group(
            self._process_group_backend, rank=global_rank, world_size=world_size, store=self.store
        )

        if self._process_group_backend == "nccl":
            atexit.register(_destroy_dist_connection)

        # On rank=0 let everyone know training is starting
        rank_zero_info(
            f"{'-' * 100}\n"
            f"distributed_backend={self._process_group_backend}\n"
            f"All distributed processes registered. Starting with {world_size} processes\n"
            f"{'-' * 100}\n"
        )

        # if 'device_mesh' in the `kwargs` is provided as a tuple, update it into the `DeviceMesh` object here
        if isinstance(self.kwargs.get("device_mesh"), tuple):
            from torch.distributed.device_mesh import init_device_mesh

            self.kwargs["device_mesh"] = init_device_mesh("cuda", self.kwargs["device_mesh"])

        init_model_parallel(self.model)

    @override
    def setup(self, trainer: pl.Trainer) -> None:
        """Connect strategy to trainer and handle adjustments before the loop starts."""
        self.trainer = trainer
        setup_data_sampler(self.trainer)
        fix_progress_bar(trainer)
        super().setup(trainer)

    def _get_loss_reduction(self, step_type: str):
        for fn_name in [f"{step_type}_loss_reduction", "loss_reduction"]:
            if hasattr(self.lightning_module, fn_name):
                return getattr(self.lightning_module, fn_name)
        return None

    def _step_proxy(self, step_type, batch, batch_idx=None):
        method_name = f"{step_type}_step"
        if self.model != self.lightning_module:
            loss = self._forward_redirection(self.model, self.lightning_module, method_name, batch, batch_idx)
        else:
            loss = getattr(self.lightning_module, method_name)(batch, batch_idx)

        _loss_reduction = self._get_loss_reduction(step_type)
        if _loss_reduction:
            return _loss_reduction.forward(batch, loss)
        return loss, {'avg': loss}

    @override
    def training_step(self, batch, batch_idx=None) -> STEP_OUTPUT:
        """Run training step and logs results."""
        assert self.lightning_module is not None
        assert self.model is not None
        with self.precision_plugin.train_step_context():
            loss, reduced = self._step_proxy("training", batch, batch_idx)

            self.lightning_module.log(
                'global_step',
                self.trainer.global_step,
                prog_bar=True,
                rank_zero_only=True,
                batch_size=1,
            )

            self.lightning_module.log(
                'step',
                self.trainer.global_step,
            )
            self.lightning_module.log(
                'reduced_train_loss', reduced['avg'], prog_bar=True, rank_zero_only=True, batch_size=1
            )

            # returns unreduced loss for backward
            return loss

    @override
    def validation_step(self, batch, batch_idx=None) -> Any:
        """Run validation step and logs results."""
        assert self.lightning_module is not None
        assert self.model is not None
        with self.precision_plugin.val_step_context():
            loss, reduced = self._step_proxy("validation", batch, batch_idx)
            self.lightning_module.log('val_loss', reduced['avg'], rank_zero_only=True, batch_size=1)
            return loss

    @override
    def test_step(self, batch, batch_idx=None) -> STEP_OUTPUT:
        """Run test step and logs results."""
        assert self.lightning_module is not None
        assert self.model is not None
        with self.precision_plugin.test_step_context():
            loss, reduced = self._step_proxy("test", batch, batch_idx)
            self.lightning_module.log('test_loss', reduced['avg'], rank_zero_only=True, batch_size=1)

            return loss

    @override
    def predict_step(self, batch, batch_idx=None) -> STEP_OUTPUT:
        """Run prediction step."""
        assert self.lightning_module is not None
        assert self.model is not None
        with self.precision_plugin.predict_step_context():
            loss, reduced = self._step_proxy("predict", batch, batch_idx)
            return reduced

    @override
    def process_dataloader(self, dataloader: DataLoader) -> DataLoader:
        """Transform dataloader with sampler."""
        if self.data_sampler:
            return self.data_sampler.transform_dataloader(dataloader)

        return dataloader

    @property
    @override
    def checkpoint_io(self) -> CheckpointIO:
        """Get CheckpointIO."""
        if not self._checkpoint_io:
            self._checkpoint_io = create_checkpoint_io()

        return self._checkpoint_io

    @checkpoint_io.setter
    def checkpoint_io(self, io: CheckpointIO) -> None:
        """Set CheckpointIO."""
        self._checkpoint_io = io

    @property
    def current_epoch_step(self) -> int:
        """
        Get the value of step within an epoch.
        """
        return max(
            self.trainer.fit_loop.epoch_loop.automatic_optimization.optim_progress.optimizer.step.current.completed,
            self.trainer.fit_loop.epoch_loop.manual_optimization.optim_step_progress.current.completed,
        )

    @override
    def remove_checkpoint(self, filepath: Union[str, Path]) -> None:
        """Delete checkpoint at filepath."""
        # Taken from MegatronStrategy
        ckpt = ckpt_to_dir(filepath)
        if self.is_global_zero:
            if os.path.islink(ckpt):
                os.unlink(ckpt)
            else:
                shutil.rmtree(ckpt)

    @override
    def save_checkpoint(
        self, checkpoint: Dict[str, Any], filepath: Union[str, Path], storage_options: Optional[Any] = None
    ) -> None:
        """Converts PyT checkpoints to MCore format and save using MCore dist ckpt library."""
        checkpoint["sharded_state_dict"] = pyt_to_mcore_state_dict(checkpoint.pop("state_dict"))
        checkpoint["state_dict"] = OrderedDict([])

        if "optimizer_states" in checkpoint and self.trainer.state.fn == TrainerFn.FITTING:
            # Clear the optimizer states. This handles the case where ckpt_save_optimizer=False
            # Ideally, the optimizer state dicts should not be generated in this case
            checkpoint["optimizer_states"] = {}

            # replace unsharded optimizer_states with sharded dict.
            # note that if trainer.save_checkpoint(path, save_weights_only=True) is called,
            # the checkpoint will contain only model weights. Optimizer states will be omitted.
            if self.ckpt_save_optimizer:
                checkpoint['optimizer'] = get_optimizer_state_dict(self.model, self.optimizers)
                pyt_to_mcore_state_dict(checkpoint['optimizer']['state'], prefix="optimizer.state.")

        self.checkpoint_io.save_checkpoint(checkpoint, filepath, storage_options=storage_options)

    @override
    def load_checkpoint(self, checkpoint_path: str | Path) -> Dict[str, Any]:
        """PTL method which we override to integrate distributed checkpoints for FSDP models.
        Different from MegatronStrategy, both model and optimizer states are restore within
        this method.

        The logic here is slightly more complicated:
        1. Obtain PyT state dicts (sharded & unflattened) for model and optim -> torch::ShardedTensor
        2. Convert to MCore state dicts -> mcore::ShardedTensor
        3. Load from checkpoint using MCore dist ckpt API -> torch::Tensor
        4. Convert to PyT state dicts (sharded & unflattened) -> torch::ShardedTensor
        5. Load into model and optim using PyT dist ckpt API
        6. Return the loaded checkpoint for lightning to load other metadata
        """
        path = Path(self.broadcast(checkpoint_path))
        torch.cuda.empty_cache()

        # TODO: the elegant way to load both state dicts. Need pytorch 2.3.1
        # msd, osd = get_state_dict(self.model, self.optimizers, options=StateDictOptions(cpu_offload=True))
        sharded_state_dict = {}
        with _get_sharded_state_dict_context(self.model):
            msd = self.model.state_dict()
            pyt_to_mcore_state_dict(msd)
            sharded_state_dict["sharded_state_dict"] = msd

        if self.ckpt_load_optimizer and self.trainer.state.fn == TrainerFn.FITTING:
            osd = get_optimizer_state_dict(self.model, self.optimizers, options=StateDictOptions(cpu_offload=True))
            pyt_to_mcore_state_dict(osd['state'], prefix="optimizer.state.")
            sharded_state_dict["optimizer"] = osd

        checkpoint = self.checkpoint_io.load_checkpoint(path, sharded_state_dict=sharded_state_dict)
        mcore_to_pyt_sharded_state_dict(checkpoint['sharded_state_dict'], msd)

        if self.ckpt_load_optimizer and self.trainer.state.fn == TrainerFn.FITTING:
            mcore_to_pyt_sharded_state_dict(checkpoint['optimizer']['state'], osd['state'])

        set_state_dict(
            self.model,
            self.optimizers if self.ckpt_load_optimizer else [],
            model_state_dict=checkpoint['sharded_state_dict'],
            optim_state_dict=checkpoint['optimizer'] if self.ckpt_load_optimizer else None,
        )

        return checkpoint



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for `StreamingBatchedAudioBuffer` and accompanying helper
classes defined in
`nemo.collections.asr.parts.utils.streaming_utils`.
"""

from __future__ import annotations

import math

import pytest
import torch

from nemo.collections.asr.parts.utils.streaming_utils import ContextSize, ContextSizeBatch, StreamingBatchedAudioBuffer

# -----------------------------------------------------------------------------
# Helper constants / fixtures
# -----------------------------------------------------------------------------

DEVICES: list[torch.device] = [torch.device("cpu")]
if torch.cuda.is_available():
    DEVICES.append(torch.device("cuda:0"))

# -----------------------------------------------------------------------------
# Tests for ContextSize and ContextSizeBatch
# -----------------------------------------------------------------------------


@pytest.mark.unit
def test_context_size_total_and_subsample():
    ctx = ContextSize(left=4, chunk=2, right=1)
    assert ctx.total() == 7

    half_ctx = ctx.subsample(factor=2)
    assert isinstance(half_ctx, ContextSize)
    assert half_ctx.left == 2 and half_ctx.chunk == 1 and half_ctx.right == 0
    assert half_ctx.total() == math.floor(7 / 2)


@pytest.mark.unit
@pytest.mark.parametrize("device", DEVICES)
def test_context_size_batch_total_and_subsample(device: torch.device):
    left = torch.tensor([4, 4], dtype=torch.long, device=device)
    chunk = torch.tensor([2, 2], dtype=torch.long, device=device)
    right = torch.tensor([2, 2], dtype=torch.long, device=device)
    batch_ctx = ContextSizeBatch(left=left, chunk=chunk, right=right)

    # total() should equal element-wise sum
    expected_total = left + chunk + right
    assert torch.equal(batch_ctx.total(), expected_total)

    # After subsampling by 2 each component should be halved (floor division)
    half_ctx = batch_ctx.subsample(2)
    assert torch.equal(half_ctx.left, left // 2)
    assert torch.equal(half_ctx.chunk, chunk // 2)
    assert torch.equal(half_ctx.right, right // 2)


# -----------------------------------------------------------------------------
# Tests for StreamingBatchedAudioBuffer
# -----------------------------------------------------------------------------


def _create_audio_batch(batch_size: int, length: int, device: torch.device, dtype: torch.dtype = torch.float32):
    """Create a dummy audio batch of shape (batch_size, length)."""
    # Use a simple ramp signal to ease debugging.
    vals = torch.arange(batch_size * length, device=device, dtype=dtype)
    return vals.view(batch_size, length)


@pytest.mark.unit
@pytest.mark.parametrize("device", DEVICES)
def test_streaming_batched_audio_buffer(device: torch.device):
    batch_size = 2
    expected_ctx = ContextSize(left=4, chunk=2, right=1)  # total = 7
    buffer = StreamingBatchedAudioBuffer(
        batch_size=batch_size,
        context_samples=expected_ctx,
        dtype=torch.float32,
        device=device,
    )

    # ------------------------------------------------------------------
    # First add : chunk + right (filling initial buffer)
    # ------------------------------------------------------------------
    first_len = expected_ctx.chunk + expected_ctx.right  # 3
    audio_batch = _create_audio_batch(batch_size, first_len, device)
    audio_lens = torch.full(
        [
            batch_size,
        ],
        first_len,
        dtype=torch.long,
        device=device,
    )
    buffer.add_audio_batch_(
        audio_batch=audio_batch,
        audio_lengths=audio_lens,
        is_last_chunk=False,
        is_last_chunk_batch=torch.zeros(batch_size, dtype=torch.bool, device=device),
    )

    # Validate context sizes
    assert buffer.context_size.left == 0
    assert buffer.context_size.chunk == expected_ctx.chunk
    assert buffer.context_size.right == expected_ctx.right
    assert buffer.samples.shape[1] == first_len  # No truncation yet

    # ------------------------------------------------------------------
    # Second add : only chunk length
    # ------------------------------------------------------------------
    chunk_len = expected_ctx.chunk  # 2
    audio_batch = _create_audio_batch(batch_size, chunk_len, device)
    audio_lens.fill_(chunk_len)
    buffer.add_audio_batch_(
        audio_batch=audio_batch,
        audio_lengths=audio_lens,
        is_last_chunk=False,
        is_last_chunk_batch=torch.zeros(batch_size, dtype=torch.bool, device=device),
    )

    # After second add, left should have grown by previous chunk (2)
    assert buffer.context_size.left == 2
    assert buffer.context_size.chunk == expected_ctx.chunk
    assert buffer.context_size.right == expected_ctx.right
    assert buffer.samples.shape[1] == 5  # 2 (left) + 2 (chunk) + 1 (right)

    # ------------------------------------------------------------------
    # Third add : another chunk, buffer should now reach full capacity (7)
    # ------------------------------------------------------------------
    buffer.add_audio_batch_(
        audio_batch=audio_batch,
        audio_lengths=audio_lens,
        is_last_chunk=False,
        is_last_chunk_batch=torch.zeros(batch_size, dtype=torch.bool, device=device),
    )

    assert buffer.samples.shape[1] == expected_ctx.total()
    assert buffer.context_size.total() == expected_ctx.total()

    # ------------------------------------------------------------------
    # Fourth add : buffer overflows by 2 samples; implementation should
    # drop the excess from the left context.
    # ------------------------------------------------------------------
    buffer.add_audio_batch_(
        audio_batch=audio_batch,
        audio_lengths=audio_lens,
        is_last_chunk=False,
        is_last_chunk_batch=torch.zeros(batch_size, dtype=torch.bool, device=device),
    )

    # Buffer length remains constant (total context size)
    assert buffer.samples.shape[1] == expected_ctx.total()
    assert buffer.context_size.total() == expected_ctx.total()

    # Left context should have been clipped by 2 samples (from 6 to 4)
    assert buffer.context_size.left == expected_ctx.left  # 4

    # ------------------------------------------------------------------
    # Final add : mark last chunk with shorter length; right context
    # should go to 0 afterwards.
    # ------------------------------------------------------------------
    last_len = 1
    audio_batch = _create_audio_batch(batch_size, last_len, device)
    audio_lens.fill_(last_len)
    buffer.add_audio_batch_(
        audio_batch=audio_batch,
        audio_lengths=audio_lens,
        is_last_chunk=True,
        is_last_chunk_batch=torch.ones(batch_size, dtype=torch.bool, device=device),
    )

    # After last chunk, right context must be zero and total size preserved
    assert buffer.context_size.right == 0
    assert buffer.context_size.total() == expected_ctx.total()
    assert buffer.samples.shape[1] == expected_ctx.total()


@pytest.mark.unit
@pytest.mark.parametrize("device", DEVICES)
def test_streaming_batched_audio_buffer_raises_on_too_long_chunk(device: torch.device):
    """`add_audio_batch_` should raise if provided chunk is larger than chunk + right."""

    expected_ctx = ContextSize(left=0, chunk=2, right=1)
    buffer = StreamingBatchedAudioBuffer(
        batch_size=1,
        context_samples=expected_ctx,
        dtype=torch.float32,
        device=device,
    )

    # Attempt to add a chunk that is too long (4 > 3)
    too_long_chunk_size = expected_ctx.chunk + expected_ctx.right + 1
    audio = _create_audio_batch(1, too_long_chunk_size, device)
    audio_lens = torch.tensor([too_long_chunk_size], dtype=torch.long, device=device)

    with pytest.raises(ValueError):
        buffer.add_audio_batch_(
            audio_batch=audio,
            audio_lengths=audio_lens,
            is_last_chunk=False,
            is_last_chunk_batch=torch.tensor([False], dtype=torch.bool, device=device),
        )



# Copyright 2024 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

import torch.nn as nn
from torch.nn import functional as F


def zero_module(module):
    """
    Initializes all parameters of the given module to zero.

    Args:
        module (nn.Module): The module whose parameters will be initialized to zero.

    Returns:
        nn.Module: The same module with zero-initialized parameters.
    """
    for p in module.parameters():
        nn.init.zeros_(p)
    return module


class ControlNetConditioningEmbedding(nn.Module):
    """
    Quoting from https://arxiv.org/abs/2302.05543: "Stable Diffusion uses a pre-processing method similar to VQ-GAN
    [11] to convert the entire dataset of 512 × 512 images into smaller 64 × 64 “latent images” for stabilized
    training. This requires ControlNets to convert image-based conditions to 64 × 64 feature space to match the
    convolution size. We use a tiny network E(·) of four convolution layers with 4 × 4 kernels and 2 × 2 strides
    (activated by ReLU, channels are 16, 32, 64, 128, initialized with Gaussian weights, trained jointly with the full
    model) to encode image-space conditions ... into feature maps ..."
    """

    def __init__(
        self,
        conditioning_embedding_channels: int,
        conditioning_channels: int = 3,
        block_out_channels: Tuple[int, ...] = (16, 32, 96, 256),
    ):
        """
        Initializes the model with convolutional layers for processing conditioning inputs.

        Args:
            conditioning_embedding_channels (int):
                Number of output channels for the conditioning embedding.
            conditioning_channels (int):
                Number of input channels for the conditioning data. Default is 3.
            block_out_channels (Tuple[int, ...]):
                Tuple specifying the output channels for each block. Default is (16, 32, 96, 256).
        """
        super().__init__()

        self.conv_in = nn.Conv2d(conditioning_channels, block_out_channels[0], kernel_size=3, padding=1)

        self.blocks = nn.ModuleList([])

        for i in range(len(block_out_channels) - 1):
            channel_in = block_out_channels[i]
            channel_out = block_out_channels[i + 1]
            self.blocks.append(nn.Conv2d(channel_in, channel_in, kernel_size=3, padding=1))
            self.blocks.append(nn.Conv2d(channel_in, channel_out, kernel_size=3, padding=1, stride=2))

        self.conv_out = zero_module(
            nn.Conv2d(block_out_channels[-1], conditioning_embedding_channels, kernel_size=3, padding=1)
        )

    def forward(self, conditioning):
        """
        Passes the conditioning input through the model to produce an embedding.

        Args:
            conditioning (torch.Tensor): Input tensor representing conditioning data.

        Returns:
            torch.Tensor: The resulting embedding tensor after processing through the network.
        """
        embedding = self.conv_in(conditioning)
        embedding = F.silu(embedding)

        for block in self.blocks:
            embedding = block(embedding)
            embedding = F.silu(embedding)

        embedding = self.conv_out(embedding)

        return embedding
